# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
)
from numpy import (
    allclose,
    array,
    log10 as log,
    ndarray,
)
from pyimpspec import (
    Circuit,
    DRTError,
    parse_cdc,
)
from pyimpspec.analysis.drt.mRQfit import _validate_circuit
from pyimpspec.analysis.fitting import _calculate_residuals
import dearpygui.dearpygui as dpg
from deareis.utility import (
    align_numbers,
    format_number,
    is_filtered_item_visible,
    render_math,
)
from deareis.data.drt import (
    DRTSettings,
    DRTResult,
)
from deareis.data.fitting import FitResult
from deareis.enums import (
    DRTMethod,
    DRTMode,
    DRTOutput,
    derivative_order_to_label,
    drt_method_to_label,
    drt_mode_to_label,
    label_to_derivative_order,
    label_to_drt_method,
    label_to_drt_mode,
    label_to_drt_output,
    label_to_rbf_shape,
    label_to_rbf_type,
    rbf_shape_to_label,
    rbf_type_to_label,
)
from deareis.gui.plots import (
    Impedance,
    DRT,
    Residuals,
)
from deareis.enums import (
    Context,
)
import deareis.tooltips as tooltips
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
from deareis.signals import Signal
import deareis.signals as signals
import deareis.themes as themes
from deareis.data import (
    DataSet,
)


MATH_DRT_WIDTH: int = 350
MATH_DRT_HEIGHT: int = 40
MATH_RBF_WIDTH: int = 380
MATH_RBF_HEIGHT: int = 40
MATH_DRT_WITHOUT_INDUCTANCE: int = -1
MATH_DRT_WITH_INDUCTANCE: int = -1
MATH_PIECEWISE_1: int = -1
MATH_PIECEWISE_2: int = -1
MATH_PIECEWISE_3: int = -1
MATH_C0_MATERN: int = -1
MATH_C2_MATERN: int = -1
MATH_C4_MATERN: int = -1
MATH_C6_MATERN: int = -1
MATH_CAUCHY: int = -1
MATH_GAUSSIAN: int = -1
MATH_INVERSE_QUADRATIC: int = -1
MATH_INVERSE_QUADRIC: int = -1
MATH_GAMMA_LN_TAU: int = -1
MATH_X: int = -1
MATH_TAU_M: int = -1
MATH_FWHM: int = -1
MATH_SHAPE: int = -1


def process_math():
    global MATH_DRT_WITHOUT_INDUCTANCE
    global MATH_DRT_WITH_INDUCTANCE
    global MATH_PIECEWISE_1
    global MATH_PIECEWISE_2
    global MATH_PIECEWISE_3
    global MATH_C0_MATERN
    global MATH_C2_MATERN
    global MATH_C4_MATERN
    global MATH_C6_MATERN
    global MATH_CAUCHY
    global MATH_GAUSSIAN
    global MATH_INVERSE_QUADRATIC
    global MATH_INVERSE_QUADRIC
    global MATH_GAMMA_LN_TAU
    global MATH_X
    global MATH_TAU_M
    global MATH_FWHM
    global MATH_SHAPE
    MATH_DRT_WITHOUT_INDUCTANCE = render_math(
        r"$Z_{\rm DRT} = R_{\infty} + \int_{-\infty}^{\infty}\ \frac{\gamma \ln{\tau}}{1 + i 2 \pi f \tau}d\ln{\tau}$",
        MATH_DRT_WIDTH,
        MATH_DRT_HEIGHT,
    )
    MATH_DRT_WITH_INDUCTANCE = render_math(
        r"$Z_{\rm DRT} = R_{\infty} + i 2 \pi f L + \int_{-\infty}^{\infty}\ \frac{\gamma \ln{\tau}}{1 + i 2 \pi f \tau}d\ln{\tau}$",
        MATH_DRT_WIDTH,
        MATH_DRT_HEIGHT,
    )
    MATH_PIECEWISE_1 = render_math(
        r"$\Phi_m(\tau) = 1-\frac{\ln{\tau} - \ln{\tau_m}}{\ln{\tau_{m-1} - \ln{\tau}}},\ \tau_{m-1} < \tau \leq \tau_m$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_PIECEWISE_2 = render_math(
        r"$\Phi_m(\tau) = 1-\frac{\ln{\tau} - \ln{\tau_m}}{\ln{\tau_{m+1} - \ln{\tau_m}}},\ \tau_m < \tau \leq \tau_{m+1}$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_PIECEWISE_3 = render_math(
        r"$\Phi_m(\tau) = 0,\ \tau_{m-1} < \tau\ {\rm or}\ \tau_{m+1} > \tau$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_C0_MATERN = render_math(
        r"$\Phi_\mu(x) = \exp(-|\mu x|)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_C2_MATERN = render_math(
        r"$\Phi_\mu(x) = \exp(-|\mu x|)(1+|\mu x|)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_C4_MATERN = render_math(
        r"$\Phi_\mu(x) = \exp(-|\mu x|)(1+|\mu x|+\frac{1}{3}|\mu x|^2)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_C6_MATERN = render_math(
        r"$\Phi_\mu(x) = \exp(-|\mu x|)(1+|\mu x|+\frac{2}{5}|\mu x|^2+\frac{1}{15}|\mu x|^3)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_CAUCHY = render_math(
        r"$\Phi_\mu(x) = 1/(1+|\mu x|)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_GAUSSIAN = render_math(
        r"$\Phi_\mu(x) = \exp(-(\mu x)^2)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_INVERSE_QUADRATIC = render_math(
        r"$\Phi_\mu(x) = 1/(1+(\mu x)^2)$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_INVERSE_QUADRIC = render_math(
        r"$\Phi_\mu(x) = 1/\sqrt{1+(\mu x)^2}$",
        MATH_RBF_WIDTH,
        MATH_RBF_HEIGHT,
    )
    MATH_GAMMA_LN_TAU = render_math(r"$\gamma(\ln{\tau})$", 44, 20, fontsize=10)
    MATH_X = render_math(r"$x = |\ln{\tau} - \ln{\tau_m}|$", 96, 20, fontsize=10)
    MATH_TAU_M = render_math(r"$\tau_m$", 18, 20, fontsize=10)
    MATH_FWHM = render_math(r"$\mu = \frac{\Delta \ln{\tau}}{k}$", 200, 40)
    MATH_SHAPE = render_math(r"$\mu = k$", 200, 40)


signals.register(Signal.RENDER_MATH, process_math)


class SettingsMenu:
    def __init__(self, default_settings: DRTSettings, label_pad: int):
        self.main_group: int = dpg.generate_uuid()
        with dpg.child_window(border=False, height=250):
            with dpg.group(horizontal=True):
                dpg.add_text("Method".rjust(label_pad))
                attach_tooltip(tooltips.drt.method)
                self.method_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=drt_method_to_label[default_settings.method],
                    items=list(label_to_drt_method.keys()),
                    width=-1,
                    callback=lambda s, a, u: self.update_settings(),
                    tag=self.method_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Mode".rjust(label_pad))
                attach_tooltip(tooltips.drt.mode)
                self.mode_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=drt_mode_to_label[default_settings.mode],
                    items=list(label_to_drt_mode.keys()),
                    width=-1,
                    callback=lambda s, a, u: self.update_settings(),
                    tag=self.mode_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Lambda".rjust(label_pad))
                attach_tooltip(tooltips.drt.lambda_value)
                self.lambda_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.lambda_value <= 0.0,
                    callback=lambda s, a, u: self.update_settings(),
                    tag=self.lambda_checkbox,
                )
                self.lambda_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    default_value=default_settings.lambda_value,
                    width=-1,
                    min_value=1e-16,
                    min_clamped=True,
                    step=0.0,
                    format="%.3g",
                    on_enter=True,
                    tag=self.lambda_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Derivative order".rjust(label_pad))
                with dpg.tooltip(dpg.last_item()):
                    with dpg.group(horizontal=True):
                        dpg.add_text("The derivative order of")
                        dpg.add_image(MATH_GAMMA_LN_TAU)
                        dpg.add_text(
                            "to use as the penalty in the Tikhonov regularization."
                        )
                    dpg.add_text(
                        "\nThis is only used when the method setting is set to BHT or TR-RBF."
                    )
                self.derivative_order_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=derivative_order_to_label[
                        default_settings.derivative_order
                    ],
                    items=list(label_to_derivative_order.keys()),
                    width=-1,
                    tag=self.derivative_order_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("RBF type".rjust(label_pad))
                with dpg.tooltip(dpg.last_item()):
                    dpg.add_text(
                        "The type of radial basis function to use with the BHT and TR-RBF methods."
                    )
                    with dpg.table(
                        borders_outerV=True,
                        borders_outerH=True,
                        borders_innerV=True,
                        borders_innerH=True,
                        scrollY=True,
                        freeze_rows=1,
                        height=18 + 8 * (MATH_RBF_HEIGHT + 4),
                    ):
                        dpg.add_table_column(
                            label="Label",
                            width_fixed=True,
                        )
                        dpg.add_table_column(
                            label="Equation",
                            width_fixed=False,
                        )
                        with dpg.table_row():
                            dpg.add_text("C^0 Matérn")
                            dpg.add_image(MATH_C0_MATERN)
                        with dpg.table_row():
                            dpg.add_text("C^2 Matérn")
                            dpg.add_image(MATH_C2_MATERN)
                        with dpg.table_row():
                            dpg.add_text("C^4 Matérn")
                            dpg.add_image(MATH_C4_MATERN)
                        with dpg.table_row():
                            dpg.add_text("C^6 Matérn")
                            dpg.add_image(MATH_C6_MATERN)
                        with dpg.table_row():
                            dpg.add_text("Cauchy")
                            dpg.add_image(MATH_CAUCHY)
                        with dpg.table_row():
                            dpg.add_text("Gaussian")
                            dpg.add_image(MATH_GAUSSIAN)
                        with dpg.table_row():
                            dpg.add_text("Inverse quadratic")
                            dpg.add_image(MATH_INVERSE_QUADRATIC)
                        with dpg.table_row():
                            dpg.add_text("Inverse quadric")
                            dpg.add_image(MATH_INVERSE_QUADRIC)
                    with dpg.group(horizontal=True):
                        dpg.add_text("µ is the shape coefficient and")
                        dpg.add_image(MATH_X)
                        dpg.add_text("where")
                        dpg.add_image(MATH_TAU_M)
                        dpg.add_text("is the mth collocation point.")
                    dpg.add_text(
                        "Alternatively, piecewise linear discretization can also be used:"
                    )
                    dpg.add_image(MATH_PIECEWISE_1)
                    dpg.add_image(MATH_PIECEWISE_2)
                    dpg.add_image(MATH_PIECEWISE_3)
                self.rbf_type_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=rbf_type_to_label[default_settings.rbf_type],
                    items=list(label_to_rbf_type.keys()),
                    width=-1,
                    tag=self.rbf_type_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Shape type".rjust(label_pad))
                with dpg.tooltip(dpg.last_item()):
                    dpg.add_text(
                        "This is only used when the method setting is set to BHT or TR-RBF."
                    )
                    with dpg.table(
                        borders_outerV=True,
                        borders_outerH=True,
                        borders_innerV=True,
                        borders_innerH=True,
                        scrollY=True,
                        freeze_rows=1,
                        height=18 + 2 * (40 + 4),
                    ):
                        dpg.add_table_column(
                            label="Label",
                            width_fixed=True,
                        )
                        dpg.add_table_column(
                            label="Equation",
                            width_fixed=False,
                        )
                        with dpg.table_row():
                            dpg.add_text("FWHM")
                            dpg.add_image(MATH_FWHM)
                        with dpg.table_row():
                            dpg.add_text("Factor")
                            dpg.add_image(MATH_SHAPE)
                    dpg.add_text(
                        "µ is used in the setting above, k is the input value, and FWHM stands for full width half maximum."
                    )
                self.rbf_shape_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=rbf_shape_to_label[default_settings.rbf_shape],
                    items=list(label_to_rbf_shape.keys()),
                    width=-1,
                    tag=self.rbf_shape_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Shape coefficient".rjust(label_pad))
                attach_tooltip(tooltips.drt.shape_coeff)
                self.shape_coeff_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    default_value=default_settings.shape_coeff,
                    width=-1,
                    min_value=1e-12,
                    min_clamped=True,
                    step=0.0,
                    format="%.3g",
                    on_enter=True,
                    tag=self.shape_coeff_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Include inductance".rjust(label_pad))
                with dpg.tooltip(dpg.last_item()):
                    dpg.add_text(
                        "Whether or not to include an inductive term in the calculations."
                    )
                    with dpg.table(
                        borders_outerV=True,
                        borders_outerH=True,
                        borders_innerV=True,
                        borders_innerH=True,
                        scrollY=True,
                        freeze_rows=1,
                        height=18 + 2 * (MATH_DRT_HEIGHT + 4),
                    ):
                        dpg.add_table_column(
                            label="State",
                            width_fixed=True,
                        )
                        dpg.add_table_column(
                            label="Equation",
                            width_fixed=False,
                        )
                        with dpg.table_row():
                            dpg.add_text("True")
                            dpg.add_image(MATH_DRT_WITH_INDUCTANCE)
                        with dpg.table_row():
                            dpg.add_text("False")
                            dpg.add_image(MATH_DRT_WITHOUT_INDUCTANCE)
                    dpg.add_text(
                        "This is only used when the method setting is set to TR-RBF."
                    )
                self.inductance_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.inductance,
                    tag=self.inductance_checkbox,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Credible intervals".rjust(label_pad))
                attach_tooltip(tooltips.drt.credible_intervals)
                self.credible_intervals_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.credible_intervals,
                    callback=lambda s, a, u: self.update_settings(),
                    tag=self.credible_intervals_checkbox,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Number of samples".rjust(label_pad))
                attach_tooltip(tooltips.drt.num_samples)
                self.num_samples_input: int = dpg.generate_uuid()
                dpg.add_input_int(
                    default_value=default_settings.num_samples,
                    width=-1,
                    min_value=1000,
                    min_clamped=True,
                    step=0,
                    on_enter=True,
                    tag=self.num_samples_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Number of attempts".rjust(label_pad))
                attach_tooltip(tooltips.drt.num_attempts)
                self.num_attempts_input: int = dpg.generate_uuid()
                dpg.add_input_int(
                    default_value=default_settings.num_attempts,
                    width=-1,
                    min_value=0,
                    min_clamped=True,
                    step=0,
                    on_enter=True,
                    tag=self.num_attempts_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Maximum symmetry".rjust(label_pad))
                attach_tooltip(tooltips.drt.maximum_symmetry)
                self.maximum_symmetry_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    default_value=default_settings.maximum_symmetry,
                    width=-1,
                    min_value=0.0,
                    max_value=1.0,
                    step=0.0,
                    format="%.3g",
                    on_enter=True,
                    tag=self.maximum_symmetry_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Circuit".rjust(label_pad))
                attach_tooltip(tooltips.drt.circuit)
                self.circuit_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value="",
                    items=[],
                    user_data={},
                    width=-1,
                    callback=lambda s, a, u: self.update_settings(),
                    tag=self.circuit_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("W".rjust(label_pad))
                attach_tooltip(tooltips.drt.W)
                self.W_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    default_value=default_settings.W,
                    width=-1,
                    min_value=0.0,
                    max_value=1.0,
                    step=0.0,
                    format="%.3g",
                    on_enter=True,
                    tag=self.W_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Num. points per decade".rjust(label_pad))
                attach_tooltip(tooltips.drt.num_per_decade)
                self.num_per_decade_input: int = dpg.generate_uuid()
                dpg.add_input_int(
                    default_value=default_settings.num_per_decade,
                    width=-1,
                    min_value=1,
                    step=0,
                    on_enter=True,
                    tag=self.num_per_decade_input,
                )
        self.update_settings()

    def update_valid_circuits(self, fits: Dict[str, FitResult]):
        lookup: Dict[str, str] = {}
        label: str
        fit: FitResult
        for label, fit in fits.items():
            try:
                _validate_circuit(fit.circuit)
            except DRTError:
                continue
            lookup[label] = fit.circuit.to_string(12)
        if len(lookup) > 0:
            longest_cdc: int = max(
                map(
                    lambda _: len(_[: _.find(" (") + 1].strip()),
                    lookup.keys(),
                )
            )
            for label in list(lookup.keys()):
                new_label: str = f"{label[:label.find(' (') + 1].strip().ljust(longest_cdc + 1)}{label[label.find(' (') + 1:]}"
                lookup[new_label] = lookup[label]
                if new_label != label:
                    del lookup[label]
        labels: List[str] = list(lookup.keys())
        dpg.configure_item(
            self.circuit_combo,
            default_value=labels[0] if len(labels) > 0 else "",
            items=labels,
            user_data=lookup,
        )
        self.update_settings()

    def get_settings(self) -> DRTSettings:
        cdc: str = dpg.get_item_user_data(self.circuit_combo).get(
            dpg.get_value(self.circuit_combo),
            "",
        )
        return DRTSettings(
            method=label_to_drt_method[dpg.get_value(self.method_combo)],
            mode=label_to_drt_mode[dpg.get_value(self.mode_combo)],
            lambda_value=dpg.get_value(self.lambda_input)
            if not dpg.get_value(self.lambda_checkbox)
            else -1.0,
            rbf_type=label_to_rbf_type[dpg.get_value(self.rbf_type_combo)],
            derivative_order=label_to_derivative_order[
                dpg.get_value(self.derivative_order_combo)
            ],
            rbf_shape=label_to_rbf_shape[dpg.get_value(self.rbf_shape_combo)],
            shape_coeff=dpg.get_value(self.shape_coeff_input),
            inductance=dpg.get_value(self.inductance_checkbox),
            credible_intervals=dpg.get_value(self.credible_intervals_checkbox),
            num_samples=dpg.get_value(self.num_samples_input),
            num_attempts=dpg.get_value(self.num_attempts_input),
            maximum_symmetry=dpg.get_value(self.maximum_symmetry_input),
            circuit=parse_cdc(cdc) if cdc != "" else None,
            W=dpg.get_value(self.W_input),
            num_per_decade=dpg.get_value(self.num_per_decade_input),
        )

    def set_settings(self, settings: DRTSettings):
        dpg.set_value(self.method_combo, drt_method_to_label[settings.method])
        self.update_settings()
        dpg.set_value(self.mode_combo, drt_mode_to_label[settings.mode])
        dpg.set_value(self.lambda_checkbox, settings.lambda_value <= 0.0)
        if settings.lambda_value > 0.0:
            dpg.set_value(self.lambda_input, settings.lambda_value)
        dpg.set_value(self.rbf_type_combo, rbf_type_to_label[settings.rbf_type])
        dpg.set_value(
            self.derivative_order_combo,
            derivative_order_to_label[settings.derivative_order],
        )
        dpg.set_value(self.rbf_shape_combo, rbf_shape_to_label[settings.rbf_shape])
        dpg.set_value(self.shape_coeff_input, settings.shape_coeff)
        dpg.set_value(self.inductance_checkbox, settings.inductance)
        dpg.set_value(self.credible_intervals_checkbox, settings.credible_intervals)
        dpg.set_value(self.num_samples_input, settings.num_samples)
        dpg.set_value(self.num_attempts_input, settings.num_attempts)
        dpg.set_value(self.maximum_symmetry_input, settings.maximum_symmetry)
        labels: List[str] = list(dpg.get_item_user_data(self.circuit_combo).keys())
        default_value: str = ""
        if settings.circuit is not None:
            cdc: str = settings.circuit.to_string(12)
            label: str
            circuit: Circuit
            for label, circuit in dpg.get_item_user_data(self.circuit_combo).items():
                if circuit == cdc:
                    default_value = label
                    break
            if default_value == "" and len(labels) > 0:
                default_value = labels[0]
        dpg.configure_item(
            self.circuit_combo,
            default_value=default_value,
        )
        dpg.set_value(self.W_input, settings.W)
        dpg.set_value(self.num_per_decade_input, settings.num_per_decade)
        self.update_settings(settings)

    def show_setting(self, widget: int):
        assert type(widget) is int and widget > 0, widget
        dpg.enable_item(widget)
        dpg.show_item(dpg.get_item_parent(widget))

    def hide_setting(self, widget: int):
        assert type(widget) is int and widget > 0, widget
        dpg.disable_item(widget)
        dpg.hide_item(dpg.get_item_parent(widget))

    def update_settings(self, settings: Optional[DRTSettings] = None):
        if settings is None:
            settings = self.get_settings()
        if settings.method == DRTMethod.TR_NNLS:
            dpg.configure_item(
                self.mode_combo,
                default_value=drt_mode_to_label[
                    DRTMode.REAL if settings.mode == DRTMode.COMPLEX else settings.mode
                ],
                items=[
                    v
                    for k, v in drt_mode_to_label.items()
                    if k == DRTMode.REAL or k == DRTMode.IMAGINARY
                ],
            )
            self.show_setting(self.mode_combo)
        elif settings.method == DRTMethod.TR_RBF:
            dpg.configure_item(
                self.mode_combo,
                default_value=drt_mode_to_label[settings.mode],
                items=list(label_to_drt_mode.keys()),
            )
            self.show_setting(self.mode_combo)
        elif settings.method == DRTMethod.BHT or settings.method == DRTMethod.M_RQ_FIT:
            self.hide_setting(self.mode_combo)
        if settings.method == DRTMethod.TR_RBF or settings.method == DRTMethod.TR_NNLS:
            self.show_setting(self.lambda_checkbox)
            if dpg.get_value(self.lambda_checkbox):
                dpg.disable_item(self.lambda_input)
            else:
                dpg.enable_item(self.lambda_input)
        else:
            self.hide_setting(self.lambda_checkbox)
            dpg.disable_item(self.lambda_input)
        if settings.method == DRTMethod.TR_RBF or settings.method == DRTMethod.BHT:
            self.show_setting(self.rbf_type_combo)
            self.show_setting(self.derivative_order_combo)
            self.show_setting(self.rbf_shape_combo)
            self.show_setting(self.shape_coeff_input)
        else:
            self.hide_setting(self.rbf_type_combo)
            self.hide_setting(self.derivative_order_combo)
            self.hide_setting(self.rbf_shape_combo)
            self.hide_setting(self.shape_coeff_input)
        if settings.method == DRTMethod.TR_RBF:
            self.show_setting(self.inductance_checkbox)
        else:
            self.hide_setting(self.inductance_checkbox)
        if settings.method == DRTMethod.TR_RBF:
            self.show_setting(self.credible_intervals_checkbox)
        else:
            self.hide_setting(self.credible_intervals_checkbox)
        if settings.method == DRTMethod.BHT:
            self.show_setting(self.num_samples_input)
        elif settings.method == DRTMethod.TR_RBF:
            self.show_setting(self.num_samples_input)
            if settings.credible_intervals is False:
                dpg.disable_item(self.num_samples_input)
        else:
            self.hide_setting(self.num_samples_input)
        if settings.method == DRTMethod.BHT:
            self.show_setting(self.num_attempts_input)
        else:
            self.hide_setting(self.num_attempts_input)
        if settings.method == DRTMethod.TR_RBF or settings.method == DRTMethod.BHT:
            self.show_setting(self.maximum_symmetry_input)
        else:
            self.hide_setting(self.maximum_symmetry_input)
        if settings.method == DRTMethod.M_RQ_FIT:
            self.show_setting(self.circuit_combo)
            self.show_setting(self.W_input)
            self.show_setting(self.num_per_decade_input)
        else:
            self.hide_setting(self.circuit_combo)
            self.hide_setting(self.W_input)
            self.hide_setting(self.num_per_decade_input)

    def has_active_input(self) -> bool:
        inputs: List[int] = [
            self.lambda_input,
            self.shape_coeff_input,
            self.num_samples_input,
            self.num_attempts_input,
            self.maximum_symmetry_input,
        ]
        tag: int
        for tag in inputs:
            if dpg.is_item_active(tag):
                return True
        return False


class StatisticsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Statistics", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 9,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=False,
                )
                for label, tooltip, filter_key in [
                    (
                        "log X²",
                        tooltips.fitting.chisqr,
                        ",".join(drt_method_to_label.values()),
                    ),
                    (
                        "Lambda",
                        tooltips.drt.lambda_value,
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.TR_NNLS,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                ]:
                    with dpg.table_row(filter_key=filter_key):
                        dpg.add_text(label.rjust(label_pad))
                        attach_tooltip(tooltip)
                        dpg.add_text("")
            dpg.add_spacer(height=8)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            cell: int = dpg.get_item_children(row, slot=1)[1]
            dpg.set_value(cell, "")

    def populate(self, drt: DRTResult):
        dpg.show_item(self._header)
        filter_key: str = drt_method_to_label[drt.settings.method]
        dpg.set_value(self._table, filter_key)
        statistics: List[str] = [
            f"{log(drt.chisqr):.3g}",
            f"{drt.lambda_value:.3e}",
        ]
        visible_items: List[bool] = []
        i: int
        row: int
        for i, row in enumerate(dpg.get_item_children(self._table, slot=1)):
            cell: int = dpg.get_item_children(row, slot=1)[2]
            dpg.set_value(cell, statistics[i])
            visible_items.append(is_filtered_item_visible(row, filter_key))
        dpg.set_item_height(
            self._table,
            18 + 23 * len(list(filter(lambda _: _ is True, visible_items))),
        )


class ScoresTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Scores", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 9,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Score".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Real (%)",
                    width_fixed=False,
                )
                dpg.add_table_column(
                    label="Imag. (%)",
                    width_fixed=False,
                )
                label: str
                tooltip: str
                for (label, tooltip) in [
                    (
                        "Mean",
                        tooltips.drt.score_mean,
                    ),
                    (
                        "Residuals, 1 sigma",
                        tooltips.drt.score_residuals,
                    ),
                    (
                        "Residuals, 2 sigma",
                        tooltips.drt.score_residuals,
                    ),
                    (
                        "Residuals, 3 sigma",
                        tooltips.drt.score_residuals,
                    ),
                    (
                        "Hellinger distance",
                        tooltips.drt.score_hd,
                    ),
                    (
                        "Jensen-Shannon distance",
                        tooltips.drt.score_jsd,
                    ),
                ]:
                    with dpg.table_row():
                        dpg.add_text(label.rjust(label_pad))
                        attach_tooltip(tooltip)
                        dpg.add_text("")  # Real
                        dpg.add_text("")  # Imag.
            dpg.add_spacer(height=8)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            cell: int
            for cell in dpg.get_item_children(row, slot=1)[2:]:
                dpg.set_value(cell, "")

    def populate(self, drt: DRTResult):
        scores: Dict[str, complex] = drt.get_scores()
        if len(scores) > 0:
            dpg.show_item(self._header)
        else:
            dpg.hide_item(self._header)
            return
        keys: List[str] = [
            "mean",
            "residuals_1sigma",
            "residuals_2sigma",
            "residuals_3sigma",
            "hellinger_distance",
            "jensen_shannon_distance",
        ]
        real_values: List[str] = []
        imaginary_values: List[str] = []
        number_format: dict = {
            "significants": 3,
            "exponent": False,
        }
        for key in keys:
            real_values.append(format_number(scores[key].real * 100, **number_format))
            imaginary_values.append(
                format_number(scores[key].imag * 100, **number_format)
            )
        real_values = align_numbers(real_values)
        imaginary_values = align_numbers(imaginary_values)
        i: int
        row: int
        for i, row in enumerate(dpg.get_item_children(self._table, slot=1)):
            tags: List[int] = dpg.get_item_children(row, slot=1)[2:]
            dpg.set_value(tags[0], real_values[i])
            dpg.set_value(tags[1], imaginary_values[i])
        dpg.set_item_height(self._table, 18 + 23 * len(keys))


class SettingsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Settings", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 12,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                for (label, filter_key) in [
                    (
                        "Method",
                        ",".join(drt_method_to_label.values()),
                    ),
                    (
                        "Mode",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.TR_NNLS,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "Lambda",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.TR_NNLS,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "RBF type",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "Deriv. order",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "RBF shape",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "Shape coeff.",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "Include inductance",
                        drt_method_to_label[DRTMethod.TR_RBF],
                    ),
                    (
                        "Credible intervals",
                        drt_method_to_label[DRTMethod.TR_RBF],
                    ),
                    (
                        "Number of samples",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                    (
                        "Number of attempts",
                        drt_method_to_label[DRTMethod.BHT],
                    ),
                    (
                        "Maximum symmetry",
                        ",".join(
                            map(
                                lambda _: drt_method_to_label[_],
                                [
                                    DRTMethod.BHT,
                                    DRTMethod.TR_RBF,
                                ],
                            )
                        ),
                    ),
                ]:
                    with dpg.table_row(filter_key=filter_key):
                        dpg.add_text(label.rjust(label_pad))
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)
            with dpg.group(horizontal=True):
                self._apply_settings_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply settings",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_DRT_SETTINGS,
                        **u,
                    ),
                    tag=self._apply_settings_button,
                )
                attach_tooltip(tooltips.general.apply_settings)
                self._apply_mask_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply mask",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        **u,
                    ),
                    tag=self._apply_mask_button,
                )
                attach_tooltip(tooltips.general.apply_mask)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[1]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, drt: DRTResult, data: DataSet):
        dpg.show_item(self._header)
        filter_key: str = drt_method_to_label[drt.settings.method]
        dpg.set_value(self._table, filter_key)
        values: List[str] = [
            drt_method_to_label[drt.settings.method],
            drt_mode_to_label[drt.settings.mode],
            f"{drt.settings.lambda_value:.3e}"
            if drt.settings.lambda_value > 0.0
            else "Automatic",
            rbf_type_to_label[drt.settings.rbf_type],
            derivative_order_to_label[drt.settings.derivative_order],
            rbf_shape_to_label[drt.settings.rbf_shape],
            f"{drt.settings.shape_coeff:.3g}",
            "True" if drt.settings.inductance else "False",
            "True" if drt.settings.credible_intervals else "False",
            str(drt.settings.num_samples),
            str(drt.settings.num_attempts),
            f"{drt.settings.maximum_symmetry:.3g}",
        ]
        visible_items: List[bool] = []
        i: int
        row: int
        for i, row in enumerate(dpg.get_item_children(self._table, slot=1)):
            if not is_filtered_item_visible(row, filter_key):
                visible_items.append(False)
                continue
            tag: List[int] = dpg.get_item_children(row, slot=1)[1]
            value: str = values[i]
            dpg.set_value(tag, value)
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))
            visible_items.append(True)
        dpg.set_item_height(
            self._table,
            18 + 23 * len(list(filter(lambda _: _ is True, visible_items))),
        )
        dpg.set_item_user_data(
            self._apply_settings_button,
            {"settings": drt.settings},
        )
        dpg.set_item_user_data(
            self._apply_mask_button,
            {
                "data": data,
                "mask": drt.mask,
                "drt": drt,
            },
        )


class DataSetsCombo:
    def __init__(self, label: str, width: int):
        self.labels: List[str] = []
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_DATA_SET,
                data=u.get(a),
            ),
            user_data={},
            width=width,
            tag=self.tag,
        )

    def populate(self, labels: List[str], lookup: Dict[str, DataSet]):
        self.labels.clear()
        self.labels.extend(labels)
        label: str = dpg.get_value(self.tag) or ""
        if labels and label not in labels:
            label = labels[0]
        dpg.configure_item(
            self.tag,
            default_value=label,
            items=labels,
            user_data=lookup,
        )

    def get(self) -> Optional[DataSet]:
        return dpg.get_item_user_data(self.tag).get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            self.labels,
        )
        dpg.set_value(self.tag, label)

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )


class ResultsCombo:
    def __init__(self, label: str, width: int):
        self.labels: Dict[str, str] = {}
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_DRT_RESULT,
                drt=u[0].get(a),
                data=u[1],
            ),
            user_data=(
                {},
                None,
            ),
            width=width,
            tag=self.tag,
        )

    def populate(self, lookup: Dict[str, DRTResult], data: Optional[DataSet]):
        self.labels.clear()
        labels: List[str] = list(lookup.keys())
        longest_cdc: int = max(list(map(lambda _: len(_[: _.find(" ")]), labels)) + [1])
        old_key: str
        for old_key in labels:
            drt: DRTResult = lookup[old_key]
            del lookup[old_key]
            cdc, timestamp = (
                old_key[: old_key.find(" ")],
                old_key[old_key.find(" ") + 1 :],
            )
            new_key: str = f"{cdc.ljust(longest_cdc)} {timestamp}"
            self.labels[old_key] = new_key
            lookup[new_key] = drt
        labels = list(lookup.keys())
        dpg.configure_item(
            self.tag,
            default_value=labels[0] if labels else "",
            items=labels,
            user_data=(
                lookup,
                data,
            ),
        )

    def get(self) -> Optional[DRTResult]:
        return dpg.get_item_user_data(self.tag)[0].get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            list(self.labels.keys()),
        )
        dpg.set_value(self.tag, self.labels[label])

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )

    def get_next_result(self) -> Optional[DRTResult]:
        lookup: Dict[str, DRTResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_result(self) -> Optional[DRTResult]:
        lookup: Dict[str, DRTResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) - 1
        return lookup[labels[index % len(labels)]]


class DRTTab:
    def __init__(self, state):
        self.state = state
        self.queued_update: Optional[Callable] = None
        label_pad: int = 24
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="DRT analysis", tag=self.tab):
            with dpg.group(horizontal=True):
                self.sidebar_width: int = 350
                self.sidebar_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=self.sidebar_width,
                    tag=self.sidebar_window,
                ):
                    # Settings
                    with dpg.child_window(
                        border=True,
                        width=-1,
                        height=290,
                    ):
                        self.settings_menu: SettingsMenu = SettingsMenu(
                            state.config.default_drt_settings, label_pad
                        )
                        with dpg.group(horizontal=True):
                            self.visibility_item: int = dpg.generate_uuid()
                            dpg.add_text("?".rjust(label_pad), tag=self.visibility_item)
                            attach_tooltip(tooltips.drt.perform)
                            self.perform_drt_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Perform analysis",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.PERFORM_DRT,
                                    data=u,
                                    settings=self.get_settings(),
                                ),
                                user_data=None,
                                width=-1,
                                tag=self.perform_drt_button,
                            )
                    # Results
                    with dpg.child_window(width=-1, height=82):
                        label_pad = 8
                        with dpg.group(horizontal=True):
                            self.data_sets_combo: DataSetsCombo = DataSetsCombo(
                                label="Data set".rjust(label_pad),
                                width=-60,
                            )
                        with dpg.group(horizontal=True):
                            self.results_combo: ResultsCombo = ResultsCombo(
                                label="Result".rjust(label_pad),
                                width=-60,
                            )
                            self.delete_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Delete",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.DELETE_DRT_RESULT,
                                    **u,
                                ),
                                user_data={},
                                width=-1,
                                tag=self.delete_button,
                            )
                            attach_tooltip(tooltips.drt.delete)
                        with dpg.group(horizontal=True):
                            dpg.add_text("Output".rjust(label_pad))
                            # TODO: Split into combo class?
                            self.output_combo: int = dpg.generate_uuid()
                            dpg.add_combo(
                                default_value=list(label_to_drt_output.keys())[0],
                                items=list(label_to_drt_output.keys()),
                                tag=self.output_combo,
                                width=-60,
                            )
                            self.copy_output_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Copy",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.COPY_OUTPUT,
                                    output=self.get_active_output(),
                                    **u,
                                ),
                                user_data={},
                                width=-1,
                                tag=self.copy_output_button,
                            )
                            attach_tooltip(tooltips.general.copy_output)
                    # Results/settings tables
                    with dpg.child_window(width=-1, height=-1):
                        self.result_group: int = dpg.generate_uuid()
                        with dpg.group(tag=self.result_group):
                            with dpg.group(show=False):
                                self.validity_text: int = dpg.generate_uuid()
                                dpg.bind_item_theme(
                                    dpg.add_text(
                                        "",
                                        wrap=self.sidebar_width - 24,
                                        tag=self.validity_text,
                                    ),
                                    themes.result.invalid,
                                )
                                dpg.add_spacer(height=8)
                            self.statistics_table: StatisticsTable = StatisticsTable()
                            self.scores_table: ScoresTable = ScoresTable()
                            self.settings_table: SettingsTable = SettingsTable()
                # Plots window
                self.plot_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=-1,
                    height=-1,
                    tag=self.plot_window,
                ):
                    self.minimum_plot_side: int = 400
                    # Gamma (or real gamma if BHT)
                    self.drt_plot: DRT = DRT(
                        width=-1,
                        height=self.minimum_plot_side,
                    )
                    self.drt_plot.plot(
                        tau=array([]),
                        gamma=array([]),
                        label="gamma",
                        theme=themes.drt.real_gamma,
                    )
                    with dpg.group(horizontal=True):
                        self.enlarge_drt_button: int = dpg.generate_uuid()
                        self.adjust_drt_limits_checkbox: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Enlarge DRT",
                            callback=self.show_enlarged_drt,
                            tag=self.enlarge_drt_button,
                        )
                        dpg.add_checkbox(
                            default_value=True,
                            tag=self.adjust_drt_limits_checkbox,
                        )
                        attach_tooltip(tooltips.general.adjust_drt_limits)
                        dpg.add_button(
                            label="Copy as CSV",
                            callback=lambda s, a, u: signals.emit(
                                Signal.COPY_PLOT_DATA,
                                plot=self.drt_plot,
                                context=Context.DRT_TAB,
                            ),
                        )
                        attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                    # Impedance plot
                    self.impedance_plot: Impedance = Impedance(
                        width=-1,
                        height=self.minimum_plot_side,
                    )
                    self.impedance_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                        labels=(
                            "Z' (d)",
                            'Z" (d)',
                        ),
                        themes=(
                            themes.impedance.real_data,
                            themes.impedance.imaginary_data,
                        ),
                    )
                    self.impedance_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                        labels=(
                            "Z' (f)",
                            'Z" (f)',
                        ),
                        fit=True,
                        themes=(
                            themes.impedance.real_simulation,
                            themes.impedance.imaginary_simulation,
                        ),
                    )
                    self.impedance_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                        labels=(
                            "Z' (f)",
                            'Z" (f)',
                        ),
                        fit=True,
                        line=True,
                        themes=(
                            themes.impedance.real_simulation,
                            themes.impedance.imaginary_simulation,
                        ),
                        show_labels=False,
                    )
                    with dpg.group(horizontal=True):
                        self.enlarge_impedance_button: int = dpg.generate_uuid()
                        self.adjust_impedance_limits_checkbox: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Enlarge impedance",
                            callback=self.show_enlarged_impedance,
                            tag=self.enlarge_impedance_button,
                        )
                        dpg.add_checkbox(
                            default_value=True,
                            tag=self.adjust_impedance_limits_checkbox,
                        )
                        attach_tooltip(tooltips.general.adjust_impedance_limits)
                        dpg.add_button(
                            label="Copy as CSV",
                            callback=lambda s, a, u: signals.emit(
                                Signal.COPY_PLOT_DATA,
                                plot=self.impedance_plot,
                                context=Context.DRT_TAB,
                            ),
                        )
                        attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                    # Residuals
                    self.residuals_plot: Residuals = Residuals(
                        width=-1,
                        height=300,
                    )
                    self.residuals_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                    )
                    with dpg.group(horizontal=True):
                        self.enlarge_residuals_button: int = dpg.generate_uuid()
                        self.adjust_residuals_limits_checkbox: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Enlarge residuals",
                            callback=self.show_enlarged_residuals,
                            tag=self.enlarge_residuals_button,
                        )
                        dpg.add_checkbox(
                            default_value=True,
                            tag=self.adjust_residuals_limits_checkbox,
                        )
                        attach_tooltip(tooltips.general.adjust_residuals_limits)
                        dpg.add_button(
                            label="Copy as CSV",
                            callback=lambda s, a, u: signals.emit(
                                Signal.COPY_PLOT_DATA,
                                plot=self.residuals_plot,
                                context=Context.DRT_TAB,
                            ),
                        )
                        attach_tooltip(tooltips.general.copy_plot_data_as_csv)

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def has_active_input(self) -> bool:
        return self.settings_menu.has_active_input()

    def get_settings(self) -> DRTSettings:
        return self.settings_menu.get_settings()

    def set_settings(self, settings: DRTSettings):
        self.settings_menu.set_settings(settings)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0
        assert type(height) is int and height > 0
        if not self.is_visible():
            return
        width, height = dpg.get_item_rect_size(self.plot_window)
        height = round(height / 2) - 24 * 2
        self.drt_plot.resize(-1, height)
        self.impedance_plot.resize(-1, height)

    def clear(self, hide: bool = True):
        self.data_sets_combo.clear()
        self.results_combo.clear()
        self.statistics_table.clear(hide=True)
        self.scores_table.clear(hide=True)
        self.settings_table.clear(hide=True)
        self.drt_plot.update(
            index=0,
            tau=array([]),
            gamma=array([]),
            label="gamma",
        )
        self.drt_plot.delete_series(from_index=1)
        self.impedance_plot.clear(delete=False)
        self.residuals_plot.clear(delete=False)

    def show_enlarged_drt(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.drt_plot,
            adjust_limits=dpg.get_value(self.adjust_drt_limits_checkbox),
        )

    def show_enlarged_impedance(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.impedance_plot,
            adjust_limits=dpg.get_value(self.adjust_impedance_limits_checkbox),
        )

    def show_enlarged_residuals(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.residuals_plot,
            adjust_limits=dpg.get_value(self.adjust_residuals_limits_checkbox),
        )

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        self.data_sets_combo.populate(labels, lookup)

    def populate_drts(self, lookup: Dict[str, DRTResult], data: Optional[DataSet]):
        assert type(lookup) is dict, lookup
        assert type(data) is DataSet or data is None, data
        self.results_combo.populate(lookup, data)
        dpg.hide_item(dpg.get_item_parent(self.validity_text))
        if data is not None and self.results_combo.labels:
            signals.emit(
                Signal.SELECT_DRT_RESULT,
                drt=self.results_combo.get(),
                data=data,
            )
        else:
            self.statistics_table.clear(hide=True)
            self.scores_table.clear(hide=True)
            self.settings_table.clear(hide=True)
            self.select_data_set(data)
            dpg.set_item_user_data(
                self.delete_button,
                {},
            )

    def get_next_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo.tag)) - 1
        return lookup[labels[index % len(labels)]]

    def get_next_result(self) -> Optional[DRTResult]:
        return self.results_combo.get_next_result()

    def get_previous_result(self) -> Optional[DRTResult]:
        return self.results_combo.get_previous_result()

    def populate_fits(self, fits: List[FitResult]):
        assert type(fits) is dict, fits
        self.settings_menu.update_valid_circuits(fits)

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.clear(hide=data is None)
        dpg.set_item_user_data(self.perform_drt_button, data)
        if data is None:
            return
        self.data_sets_combo.set(data.get_label())
        f: ndarray = data.get_frequency()
        Z: ndarray = data.get_impedance()
        self.impedance_plot.update(
            index=0,
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
        )

    def assert_drt_up_to_date(self, drt: DRTResult, data: DataSet):
        # Check if the number of unmasked points is the same
        Z_exp: ndarray = data.get_impedance()
        Z_drt: ndarray = drt.get_impedance()
        assert Z_exp.shape == Z_drt.shape, "The number of data points differ!"
        # Check if the masks are the same
        mask_exp: Dict[int, bool] = data.get_mask()
        mask_drt: Dict[int, bool] = {
            k: drt.mask.get(k, mask_exp.get(k, False)) for k in drt.mask
        }
        num_masked_exp: int = list(data.get_mask().values()).count(True)
        num_masked_drt: int = list(drt.mask.values()).count(True)
        assert num_masked_exp == num_masked_drt, "The masks are different sizes!"
        i: int
        for i in mask_drt.keys():
            assert (
                i in mask_exp
            ), f"The data set does not have a point at index {i + 1}!"
            assert (
                mask_exp[i] == mask_drt[i]
            ), f"The data set's mask differs at index {i + 1}!"
        # Check if the frequencies and impedances are the same
        assert allclose(
            drt.get_frequency(), data.get_frequency()
        ), "The frequencies differ!"
        real_residual: ndarray
        imaginary_residual: ndarray
        real_residual, imaginary_residual = _calculate_residuals(Z_exp, Z_drt)
        assert allclose(drt.real_residual, real_residual) and allclose(
            drt.imaginary_residual, imaginary_residual
        ), "The data set's impedances differ from what they were when the DRT analysis was performed!"

    def select_drt_result(
        self,
        drt: Optional[DRTResult],
        data: Optional[DataSet],
    ):
        assert type(drt) is DRTResult or drt is None, drt
        assert type(data) is DataSet or data is None, data
        dpg.set_item_user_data(
            self.delete_button,
            {
                "drt": drt,
                "data": data,
            },
        )
        dpg.set_item_user_data(
            self.copy_output_button,
            {
                "drt": drt,
                "data": data,
            },
        )
        if not self.is_visible():
            self.queued_update = lambda: self.select_drt_result(drt, data)
            return
        self.queued_update = None
        self.select_data_set(data)
        if drt is None or data is None:
            if dpg.get_value(self.adjust_drt_limits_checkbox):
                self.drt_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_residuals_limits_checkbox):
                self.residuals_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_impedance_limits_checkbox):
                self.impedance_plot.queue_limits_adjustment()
            return
        self.results_combo.set(drt.get_label())
        message: str
        try:
            self.assert_drt_up_to_date(drt, data)
            dpg.hide_item(dpg.get_item_parent(self.validity_text))
        except AssertionError as message:
            dpg.set_value(
                self.validity_text,
                f"DRT analysis result is not valid for the current state of the data set!\n\n{message}",
            )
            dpg.show_item(dpg.get_item_parent(self.validity_text))
        self.statistics_table.populate(drt)
        self.scores_table.populate(drt)
        self.settings_table.populate(drt, data)
        tau: ndarray
        gamma: ndarray
        tau, gamma = drt.get_drt_data()
        self.drt_plot.update(
            index=0,
            tau=tau,
            gamma=gamma,
            label="real" if drt.settings.method == DRTMethod.BHT else "gamma",
        )
        if (
            drt.settings.method == DRTMethod.TR_RBF
            and drt.settings.credible_intervals is True
        ):
            mean: ndarray
            lower: ndarray
            upper: ndarray
            tau, gamma, lower, upper = drt.get_drt_credible_intervals()
            self.drt_plot.plot(
                tau=tau,
                mean=gamma,
                label="mean",
                theme=themes.drt.mean_gamma,
            )
            self.drt_plot.plot(
                tau=tau,
                lower=lower,
                upper=upper,
                label="3-sigma CI",
                theme=themes.drt.credible_intervals,
            )
        if drt.settings.method == DRTMethod.BHT:
            tau, gamma = drt.get_drt_data(imaginary=True)
            self.drt_plot.plot(
                tau=tau,
                imaginary=gamma,
                label="imag.",
                theme=themes.drt.imaginary_gamma,
            )
        f: ndarray = drt.get_frequency()
        Z: ndarray = drt.get_impedance()
        self.impedance_plot.update(
            index=1,
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
        )
        self.impedance_plot.update(
            index=2,
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
        )
        real: ndarray
        imaginary: ndarray
        f, real, imaginary = drt.get_residual_data()
        self.residuals_plot.update(
            index=0,
            frequency=f,
            real=real,
            imaginary=imaginary,
        )
        if dpg.get_value(self.adjust_drt_limits_checkbox):
            self.drt_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_impedance_limits_checkbox):
            self.impedance_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_residuals_limits_checkbox):
            self.residuals_plot.queue_limits_adjustment()

    def get_active_output(self) -> Optional[DRTOutput]:
        return label_to_drt_output.get(dpg.get_value(self.output_combo))



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from multiprocessing import cpu_count
from typing import (
    Optional,
)
import deareis.api.fitting as api
from deareis.data import (
    DataSet,
    FitResult,
    FitSettings,
    PlotSettings,
    Project,
)
from deareis.gui import (
    ProjectTab,
)
from deareis.signals import Signal
import deareis.signals as signals
from deareis.state import STATE


def select_fit_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    fit: Optional[FitResult] = kwargs.get("fit")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or fit is None:
        return
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading fit result")
    project_tab.select_fit_result(fit, data)
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def delete_fit_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    fit: Optional[FitResult] = kwargs.get("fit")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or fit is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting fit result")
    settings: Optional[PlotSettings] = project_tab.get_active_plot()
    update_plot: bool = (
        fit.uuid in settings.series_order if settings is not None else False
    )
    project.delete_fit(data, fit)
    project_tab.populate_fits(project, data)
    if settings is not None:
        project_tab.plotting_tab.populate_fits(
            project.get_all_fits(),
            project.get_data_sets(),
            settings,
        )
    if update_plot:
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def apply_fit_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[FitSettings] = kwargs.get("settings")
    if settings is None:
        return
    project_tab.set_fit_settings(settings)


def perform_fit(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    data: Optional[DataSet] = kwargs.get("data")
    settings: Optional[FitSettings] = kwargs.get("settings")
    if data is None or settings is None:
        return
    assert data.get_num_points() > 0, "There are no data points to fit the circuit to!"
    # Prevent the GUI from becoming unresponsive or sluggish
    num_procs: int = max(2, cpu_count() - 1)
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Performing fit")
    fit: FitResult = api.fit_circuit(data=data, settings=settings, num_procs=num_procs)
    project.add_fit(data, fit)
    project_tab.populate_fits(project, data)
    project_tab.plotting_tab.populate_fits(
        project.get_all_fits(),
        project.get_data_sets(),
        project_tab.get_active_plot(),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List, Optional, Tuple, Type
from numpy import inf
from pyimpspec import Circuit, Element, ParsingError
import pyimpspec
import dearpygui.dearpygui as dpg
from deareis.gui.circuit_editor.parser import Parser, Node
import deareis.themes as themes
from deareis.tooltips import attach_tooltip, update_tooltip
import deareis.tooltips as tooltips
from deareis.keybindings import is_alt_down, is_control_down


class CircuitEditor:
    def __init__(self, window: int, callback: Callable):
        assert type(window) is int and dpg.does_item_exist(window), window
        self.parameter_inputs: List[int] = []
        self.callback: Callable = callback
        self.window: int = window
        with dpg.group(horizontal=True, parent=self.window):
            self.parameter_window: int = dpg.generate_uuid()
            with dpg.child_window(width=250, tag=self.parameter_window):
                pass
            with dpg.group():
                with dpg.group(horizontal=True):
                    dpg.add_text("CDC input")
                    attach_tooltip(tooltips.circuit_editor.cdc_input)
                    self.cdc_input: int = dpg.generate_uuid()
                    dpg.add_input_text(
                        width=-147,
                        tag=self.cdc_input,
                        callback=lambda s, a, u: self.parse_cdc(a),
                        on_enter=True,
                    )
                    dpg.add_button(
                        label="Parse",
                        width=50,
                        callback=lambda s, a, u: self.parse_cdc(
                            dpg.get_value(self.cdc_input)
                        ),
                    )
                    dpg.add_button(
                        label="Clear",
                        width=80,
                        callback=lambda s, a, u: self.parse_cdc(""),
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("  Element")
                    attach_tooltip(tooltips.circuit_editor.element_combo)
                    elements: Dict[str, Type[Element]] = {
                        _.get_description(): _
                        for _ in pyimpspec.get_elements().values()
                    }
                    items: List[str] = list(elements.keys())
                    element_combo: int = dpg.add_combo(
                        width=-147, items=items, default_value=items[0]
                    )
                    dpg.add_button(
                        label="Add",
                        width=50,
                        callback=lambda s, a, u: self.add_element_node(
                            u.get(dpg.get_value(element_combo))()
                        ),
                        user_data=elements,
                    )
                    dpg.add_button(
                        label="Add dummy",
                        width=80,
                        callback=lambda s, a, u: self.add_dummy_node(),
                    )
                    attach_tooltip(tooltips.circuit_editor.dummy_node)
                self.node_editor: int = dpg.generate_uuid()
                dpg.add_node_editor(
                    width=-1,
                    height=-72,
                    tag=self.node_editor,
                    minimap=True,
                    minimap_location=dpg.mvNodeMiniMap_Location_BottomRight,
                )
                self.node_handler: int = dpg.generate_uuid()
                with dpg.item_handler_registry(tag=self.node_handler):
                    dpg.add_item_clicked_handler(callback=self.node_clicked)
                self.parser: Parser = Parser(self.node_editor, self.node_handler)
                dpg.configure_item(
                    self.node_editor,
                    callback=self.link,
                    delink_callback=self.delink,
                )
                with dpg.group():
                    with dpg.group(horizontal=True):
                        dpg.add_text("   Basic CDC")
                        attach_tooltip(tooltips.circuit_editor.basic_cdc)
                        self.basic_cdc_output_field: int = dpg.generate_uuid()
                        dpg.add_input_text(
                            tag=self.basic_cdc_output_field, width=-150, enabled=False
                        )
                        dpg.add_button(
                            label="Copy to clipboard",
                            width=-1,
                            callback=lambda: dpg.set_clipboard_text(
                                dpg.get_value(self.basic_cdc_output_field)
                            ),
                        )
                    with dpg.group(horizontal=True):
                        dpg.add_text("Extended CDC")
                        attach_tooltip(tooltips.circuit_editor.extended_cdc)
                        self.extended_cdc_output_field: int = dpg.generate_uuid()
                        dpg.add_input_text(
                            tag=self.extended_cdc_output_field,
                            width=-150,
                            enabled=False,
                        )
                        dpg.add_button(
                            label="Copy to clipboard",
                            width=-1,
                            callback=lambda: dpg.set_clipboard_text(
                                dpg.get_value(self.extended_cdc_output_field)
                            ),
                        )
                    with dpg.group(horizontal=True):
                        dpg.add_text("      Status")
                        attach_tooltip(tooltips.circuit_editor.status)
                        self.status_field: int = dpg.generate_uuid()
                        dpg.add_input_text(
                            tag=self.status_field, width=-150, enabled=False
                        )
                        self.accept_button: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Accept circuit",
                            width=-1,
                            callback=lambda s, a, u: self.callback(u),
                            tag=self.accept_button,
                            show=False,
                        )
                        self.cancel_button: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Cancel",
                            width=-1,
                            callback=lambda s, a, u: self.callback(None),
                            tag=self.cancel_button,
                        )

    def show(self, circuit: Optional[Circuit]):
        assert type(circuit) is Circuit or circuit is None, circuit
        self.clear_parameter_window(add_info=True)
        self.parser.circuit_to_nodes(circuit)
        self.update(circuit, update_input=True)
        self.update_outputs(circuit=circuit)
        self.update_status("OK" if circuit is not None else "", False)
        self.setup_keybindings()
        dpg.show_item(self.window)

    def hide(self):
        if hasattr(self, "key_handler") and dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        dpg.hide_item(self.window)

    def setup_keybindings(self):
        def delete_callback():
            if not dpg.does_item_exist(self.window):
                return
            elif not dpg.is_item_shown(self.window):
                return
            elif not dpg.is_item_hovered(self.node_editor):
                return
            elif (
                dpg.is_item_active(self.cdc_input)
                or dpg.is_item_active(self.basic_cdc_output_field)
                or dpg.is_item_active(self.extended_cdc_output_field)
                or dpg.is_item_active(self.status_field)
                or any(map(lambda _: dpg.is_item_active(_), self.parameter_inputs))
            ):
                return
            tag: int
            link_tags: List[int] = dpg.get_selected_links(self.node_editor)
            if link_tags:
                for tag in link_tags:
                    self.delink(-1, tag)
            node_tags: List[int] = dpg.get_selected_nodes(self.node_editor)
            if node_tags:
                for tag in node_tags:
                    node: Node = self.parser.find_node(tag=tag)
                    if node == self.parser.we_node or node == self.parser.cere_node:
                        continue
                    self.delete_node(node)

        def accept_callback():
            if not (is_alt_down() or is_control_down()):
                return
            self.callback(dpg.get_item_user_data(self.accept_button))

        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(key=dpg.mvKey_Delete, callback=delete_callback)
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape, callback=lambda: self.callback(None)
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=accept_callback,
            )

    def update(self, circuit: Optional[Circuit], update_input: bool = False):
        dpg.hide_item(self.accept_button)
        dpg.show_item(self.cancel_button)
        if update_input:
            dpg.set_value(
                self.cdc_input, circuit.to_string() if circuit is not None else ""
            )
        if circuit is not None and circuit.to_string() not in ["[]", "()"]:
            dpg.hide_item(self.cancel_button)
            dpg.show_item(self.accept_button)
        dpg.set_item_user_data(self.accept_button, circuit)

    def parse_cdc(self, cdc: str):
        assert type(cdc) is str, cdc
        if cdc == "":
            self.clear_parameter_window()
        circuit: Optional[Circuit] = None
        try:
            circuit = pyimpspec.parse_cdc(cdc)
            dpg.bind_item_theme(self.cdc_input, themes.cdc.valid)
            self.update_outputs(circuit=circuit)
            self.update_status("OK", False)
        except ParsingError as err:
            dpg.bind_item_theme(self.cdc_input, themes.cdc.invalid)
            self.update_outputs(stack=[])
            self.update_status(str(err), True)
        if circuit is not None:
            self.parser.circuit_to_nodes(circuit)
        else:
            self.parser.clear_nodes()
        self.update(circuit, update_input=cdc == "")

    def update_status(self, msg: str, invalid: bool):
        dpg.set_value(self.status_field, msg)
        if invalid:
            dpg.bind_item_theme(self.status_field, themes.cdc.invalid)
        else:
            dpg.bind_item_theme(self.status_field, themes.cdc.valid)

    def update_outputs(self, circuit: Optional[Circuit] = None, stack: List[str] = []):
        assert type(circuit) is Circuit or (circuit is None and type(stack) is list), (
            circuit,
            stack,
        )
        theme: int = themes.cdc.valid
        basic_output: str = ""
        extended_output: str = ""
        if circuit is not None:
            basic_output = circuit.to_string()
            extended_output = circuit.to_string(4)
        else:
            theme = themes.cdc.invalid
            basic_output = "".join(
                map(lambda _: _[: _.find("{")] if "{" in _ else _, stack)
            )
            extended_output = "".join(stack)
        dpg.set_value(self.basic_cdc_output_field, basic_output)
        dpg.set_value(self.extended_cdc_output_field, extended_output)
        dpg.bind_item_theme(self.basic_cdc_output_field, theme)
        dpg.bind_item_theme(self.extended_cdc_output_field, theme)

    def clear_parameter_window(self, add_info: bool = False):
        self.parameter_inputs.clear()
        dpg.delete_item(self.parameter_window, children_only=True)
        if add_info:
            dpg.add_text(
                """
Nodes can be connected by left-clicking on the input/output of one node, dragging to the output/input of another node, and releasing. Connections can be deleted by left-clicking a connection while holding down the Ctrl key.

The parameters of the element represented by a node can be altered by left-clicking on the label of the node. The parameters that can be modified will then show up in the sidebar to the left. An element's label can also be modified.

Nodes can be deleted by left-clicking on the label of a node and then left-clicking on the 'Delete' button that shows up in the sidebar to the left. Alternatively, you can select a node and press the Delete key. Note that the 'WE' and 'CE+RE' nodes, which represent the terminals of the circuit, cannot be deleted.

You can pan the node editor by holding down the middle mouse button and moving the cursor.
            """.strip(),
                wrap=235,
                parent=self.parameter_window,
            )

    def node_clicked(self, sender: int, app_data):
        assert type(sender) is int
        self.clear_parameter_window()
        node: Node = self.parser.find_node(tag=app_data[1])
        tooltip_text: str
        if node.id == self.parser.we_node.id or node.id == self.parser.cere_node.id:
            with dpg.group(parent=self.parameter_window):
                dpg.add_text("Electrode(s)")
                with dpg.group(horizontal=True):
                    attach_tooltip(
                        """
The working electrode.
                        """.strip()
                        if node.id == self.parser.we_node.id
                        else """
The counter and reference electrodes.
                        """.strip(),
                        parent=dpg.add_text("?"),
                    )
                    dpg.add_text(
                        "WE" if node.id == self.parser.we_node.id else "CE+RE"
                    )
            return
        elif node.id < 0:
            with dpg.group(parent=self.parameter_window):
                with dpg.group(horizontal=True):
                    attach_tooltip(
                        tooltips.circuit_editor.dummy_node,
                        parent=dpg.add_text("?"),
                    )
                    dpg.add_text("Dummy node")
        else:
            dpg.add_button(
                label="Delete",
                width=-1,
                callback=lambda s, a, u: self.delete_node(node),
                parent=self.parameter_window,
            )
            element: Element = node.element
            with dpg.group(parent=self.parameter_window):
                with dpg.group(horizontal=True):
                    tooltip_text = element.get_extended_description()
                    attach_tooltip(tooltip_text, parent=dpg.add_text("?"))
                    label = element.get_description()
                    if len(label) > 30:
                        label = label[:27] + "..."
                    dpg.add_text(label)
                with dpg.group(horizontal=True):
                    dpg.add_text("Label")
                    default_label: str = element.get_default_label()
                    hint: str = default_label[default_label.find("_") + 1 :]

                    def update_label(sender: int, new_label: str):
                        assert type(sender) is int
                        assert type(new_label) is str
                        new_label = new_label.strip()
                        try:
                            int(new_label)
                            new_label = ""
                            dpg.set_value(sender, "")
                            return
                        except ValueError:
                            pass
                        if new_label == default_label:
                            new_label = ""
                        if new_label == "" and not dpg.get_value(sender) == new_label:
                            dpg.set_value(sender, new_label)
                        element.set_label(new_label)
                        node.set_label(element.get_label())
                        self.validate_nodes()

                    current_label: str = element.get_label()
                    current_label = current_label[current_label.find("_") + 1 :]
                    self.parameter_inputs.append(
                        dpg.add_input_text(
                            hint=hint,
                            default_value=current_label
                            if current_label != hint
                            else "",
                            width=-1,
                            callback=update_label,
                            on_enter=True,
                        )
                    )
                dpg.add_spacer(height=8)
                dpg.add_text("Parameters")
                key: str
                value: float
                for key, value in element.get_parameters().items():
                    self.node_parameter(element, key, value, 34)

    def delete_node(self, node: Node):
        assert type(node) is Node, node
        self.parser.delete_node(node)
        self.validate_nodes()
        self.clear_parameter_window()

    def add_element_node(self, element: Optional[Element]):
        assert element is None or isinstance(element, Element), element
        if element is None:
            return
        self.parser.add_element_node(element)
        self.validate_nodes()

    def add_dummy_node(self):
        self.parser.add_dummy_node()
        self.validate_nodes()

    def node_parameter(
        self, element: Element, key: str, initial_value: float, max_padding: int
    ):
        assert isinstance(element, Element)
        assert type(key) is str
        assert type(initial_value) is float
        assert type(max_padding) is int
        fixed: bool = element.is_fixed(key)
        enabled: bool
        lower_limit: float = element.get_lower_limit(key)
        upper_limit: float = element.get_upper_limit(key)
        current_value: float
        cv_input_field: int = dpg.generate_uuid()
        cv_checkbox: int = dpg.generate_uuid()
        ll_input_field: int = dpg.generate_uuid()
        ll_checkbox: int = dpg.generate_uuid()
        ul_input_field: int = dpg.generate_uuid()
        ul_checkbox: int = dpg.generate_uuid()
        label_pad: int = 14
        with dpg.collapsing_header(label=f" {key}", leaf=True):
            with dpg.group(horizontal=True):
                dpg.add_text("Initial value".rjust(label_pad))
                dpg.add_input_float(
                    default_value=initial_value,
                    step=0,
                    format="%.4g",
                    width=-48,
                    tag=cv_input_field,
                    on_enter=True,
                )
                dpg.add_checkbox(
                    default_value=fixed,
                    tag=cv_checkbox,
                )
                dpg.add_text("F")
                attach_tooltip("Fixed")
            with dpg.group(horizontal=True):
                dpg.add_text("Lower limit".rjust(label_pad))
                enabled = lower_limit != -inf
                dpg.add_input_float(
                    default_value=lower_limit,
                    step=0,
                    format="%.4g",
                    width=-48,
                    tag=ll_input_field,
                    on_enter=True,
                    readonly=not enabled,
                    enabled=enabled,
                )
                dpg.add_checkbox(
                    default_value=enabled,
                    tag=ll_checkbox,
                )
                dpg.add_text("E")
                attach_tooltip("Enabled")
            with dpg.group(horizontal=True):
                dpg.add_text("Upper limit".rjust(label_pad))
                enabled = upper_limit != inf
                dpg.add_input_float(
                    default_value=upper_limit,
                    step=0,
                    format="%.4g",
                    width=-48,
                    tag=ul_input_field,
                    on_enter=True,
                    readonly=not enabled,
                    enabled=enabled,
                )
                dpg.add_checkbox(
                    default_value=enabled,
                    tag=ul_checkbox,
                )
                dpg.add_text("E")
                attach_tooltip("Enabled")

            def reset_parameter():
                element.reset_parameters([key])
                dpg.set_value(cv_input_field, element.get_parameters()[key])
                dpg.set_value(cv_checkbox, element.is_fixed(key))
                value = element.get_lower_limit(key)
                dpg.configure_item(
                    ll_input_field,
                    default_value=value,
                    readonly=value == -inf,
                    enabled=value != -inf,
                )
                dpg.set_value(ll_checkbox, value != -inf)
                value = element.get_upper_limit(key)
                dpg.configure_item(
                    ul_input_field,
                    default_value=value,
                    readonly=value == inf,
                    enabled=value != inf,
                )
                dpg.set_value(ul_checkbox, value != inf)

            dpg.add_button(label="Reset", callback=reset_parameter)
        dpg.add_spacer(height=8)
        self.parameter_inputs.extend(
            [
                cv_input_field,
                ll_input_field,
                ul_input_field,
            ]
        )

        def set_lower_limit(sender: int, new_value: float):
            assert type(sender) is int
            if not dpg.get_value(ll_checkbox):
                new_value = -inf
            current_value = dpg.get_value(cv_input_field)
            if new_value > current_value:
                new_value = current_value
                dpg.configure_item(ll_input_field, default_value=new_value)
            element.set_lower_limit(key, new_value)
            self.validate_nodes()

        def toggle_lower_limit(sender: int, state: bool):
            assert type(sender) is int
            assert type(state) is bool
            new_value: Optional[float]
            if state:
                new_value = element.get_default_lower_limits().get(key)
                current_value: float = dpg.get_value(cv_input_field)
                if new_value is None or new_value == -inf or new_value > current_value:
                    new_value = 0.9 * current_value
            else:
                new_value = -inf
            dpg.configure_item(
                ll_input_field,
                default_value=new_value,
                readonly=not state,
                enabled=state,
            )
            element.set_lower_limit(key, new_value)
            self.validate_nodes()

        def set_upper_limit(sender: int, new_value: float):
            assert type(sender) is int
            if not dpg.get_value(ul_checkbox):
                new_value = inf
            current_value = dpg.get_value(cv_input_field)
            if new_value < current_value:
                new_value = current_value
                dpg.configure_item(ul_input_field, default_value=new_value)
            element.set_upper_limit(key, new_value)
            self.validate_nodes()

        def toggle_upper_limit(sender: int, state: bool):
            assert type(sender) is int
            assert type(state) is bool
            new_value: Optional[float]
            if state:
                new_value = element.get_default_upper_limits().get(key)
                current_value: float = dpg.get_value(cv_input_field)
                if new_value is None or new_value == inf or new_value < current_value:
                    new_value = 1.1 * current_value
            else:
                new_value = inf
            dpg.configure_item(
                ul_input_field,
                default_value=new_value,
                readonly=not state,
                enabled=state,
            )
            element.set_upper_limit(key, new_value)
            self.validate_nodes()

        def set_value(sender: int, new_value: float):
            assert type(sender) is int
            if dpg.get_value(ll_checkbox):
                lower_limit = dpg.get_value(ll_input_field)
                if lower_limit > new_value:
                    dpg.configure_item(ll_input_field, default_value=new_value)
                    element.set_lower_limit(key, new_value)
            if dpg.get_value(ul_checkbox):
                upper_limit = dpg.get_value(ul_input_field)
                if upper_limit < new_value:
                    dpg.configure_item(ul_input_field, default_value=new_value)
                    element.set_upper_limit(key, new_value)
            element.set_parameters({key: new_value})
            self.validate_nodes()

        def toggle_fixed(sender: int, state: bool):
            assert type(sender) is int
            assert type(state) is bool
            element.set_fixed(key, state)
            if dpg.get_value(ll_checkbox):
                dpg.set_value(ll_checkbox, False)
                toggle_lower_limit(ll_checkbox, False)
            if dpg.get_value(ul_checkbox):
                dpg.set_value(ul_checkbox, False)
                toggle_upper_limit(ul_checkbox, False)
            self.validate_nodes()

        dpg.set_item_callback(cv_input_field, set_value)
        dpg.set_item_callback(cv_checkbox, toggle_fixed)
        dpg.set_item_callback(ll_input_field, set_lower_limit)
        dpg.set_item_callback(ll_checkbox, toggle_lower_limit)
        dpg.set_item_callback(ul_input_field, set_upper_limit)
        dpg.set_item_callback(ul_checkbox, toggle_upper_limit)

    def validate_nodes(self):
        circuit: Optional[Circuit]
        msg: str
        stack: List[str]
        circuit, msg, stack = self.parser.generate_circuit()
        if circuit is None:
            self.update_outputs(stack=stack)
            self.update_status(msg, True)
        else:
            self.update_outputs(circuit=circuit)
            self.update_status("OK", False)
        self.update(circuit)

    def link(self, sender: int, attributes: Tuple[int, int]):
        self.parser.link(sender, attributes)
        self.validate_nodes()

    def delink(self, sender: int, link: int):
        self.parser.delink(sender, link)
        self.validate_nodes()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


plotting = SimpleNamespace(
    **{
        "dataset_checkbox": """
Whether or not this data set is included in the plot.
    """.strip(),
        "dataset_label": """
The label assigned to this data set in the 'Data sets' tab.
    """.strip(),
        "test_checkbox": """
Whether or not this test result is included in the plot.
    """.strip(),
        "drt_checkbox": """
Whether or not this DRT analysis result is included in the plot.
    """.strip(),
        "fit_checkbox": """
Whether or not this fit result is included in the plot.
    """.strip(),
        "simulation_checkbox": """
Whether or not this simulation result is included in the plot.
    """.strip(),
        "filter": """
Hide items that do not contain the string of text, which is entered into the input field to the right, in their labels. Test and fit results are considered to also contain the labels of their corresponding data sets.
    """.strip(),
        "select_all": """
Add all of the items listed above to the plot.
    """.strip(),
        "unselect_all": """
Remove all of the items listed above from the plot.
    """.strip(),
        "create": """
Create a new plot.
    """.strip(),
        "remove": """
Remove the current plot.
    """.strip(),
        "copy_appearance": """
Copy color, marker, and line settings from another plot and apply them to the active items.
    """.strip(),
        "item_type": """
The type of item.
    """.strip(),
        "label": """
The label used in the legend of the plot to identify the corresponding item.

This value defaults to the label of the item but it can be overridden by typing in a new label. Overriding the label with only blank space(s) will result in no entry in the legend but the data will still plotted.
    """.strip(),
        "appearance": """
The settings that define the appearance of the item in the plot.
    """.strip(),
        "position": """
Change the order in which items are added to the plot. This affects both the order of the items in the legend as well as the order in which markers and/or lines are plotted.

The two buttons can be used to adjust the position of the corresponding item.
    """.strip(),
        "collapse_expand_sidebar": """
Collapse/expand the sidebar.
    """.strip(),
        "copy_appearance_toggle": """
Toggle this to set whether or not this series should copy settings from the chosen source.
    """.strip(),
        "copy_appearance_labels": """
Toggle whether or not labels should be copied.
    """.strip(),
        "copy_appearance_colors": """
Toggle whether or not colors should be copied.
    """.strip(),
        "copy_appearance_markers": """
Toggle whether or not markers should be copied.
    """.strip(),
        "copy_appearance_lines": """
Toggle whether or not lines should be copied.
    """.strip(),
        "export_plot": """
Export the current plot using matplotlib.
    """.strip(),
        "export_units": """
The units used for the width and height.
    """.strip(),
        "export_width": """
The width of the figure.
    """.strip(),
        "export_height": """
The height of the figure.
    """.strip(),
        "export_dpi": """
The number of dots (or points) per inch.
    """.strip(),
        "export_result": """
The final dimensions.
    """.strip(),
        "export_preview": """
The maximum limit for either dimension in the preview. For example, if the limit is set to 1024 and the figure is 1920 px by 1080 px, then the preview will be 1024 px by 576 px.

Note that using a limit may cause the preview to look slightly or even very different compared to the final saved figure but helps to maintain acceptable performance.
However, choosing to not use a limit may cause the program to become unresponsive and use a lot of memory if the figure dimensions are very large.
    """.strip(),
        "export_min_x": """
The minimum limit for the x-axis.
    """.strip(),
        "export_max_x": """
The maximum limit for the x-axis.
    """.strip(),
        "export_min_y": """
The minimum limit for the y-axis.
    """.strip(),
        "export_max_y": """
The maximum limit for the y-axis.
    """.strip(),
        "export_title": """
Show the figure title.
    """.strip(),
        "export_legend": """
Show the figure legend.
    """.strip(),
        "export_grid": """
Show the grid.
    """.strip(),
        "export_tight": """
Use a tight layout that reduces the size of the margins.
    """.strip(),
        "export_file": """
Save the figure as a file.
    """.strip(),
        "export_num_per_decade": """
The number of points per decade in lines.
    """.strip(),
        "export_extension": """
The default extension to use when exporting a plot.
    """.strip(),
        "clear_registry": """
NOTE! Clearing the texture registry can cause DearEIS to crash on some systems. The cause is not known for certain at the moment but it may be a bug in the GUI framework and/or a certain GPU driver.

If enabled, then the texture registry is cleared to free memory. Disabling will cause memory to not be freed until DearEIS is closed, which may be an issue if a lot of plot previews are generated during a single session.
    """.strip(),
        "disable_preview": """
NOTE! If the texture registry is not allowed to be cleared, then DearEIS will not be able to free the memory used by the generated plot previews. The plot previews can be disabled completely with this setting to avoid such a memory leak.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
)
from numpy import (
    array,
    floor,
    log10 as log,
    ndarray,
)
from deareis.data import DataSet
from deareis.data.kramers_kronig import (
    TestResult,
    TestSettings,
)
from deareis.signals import Signal
from deareis.tooltips import attach_tooltip
from deareis.utility import calculate_window_position_dimensions
import deareis.signals as signals
import deareis.themes as themes
import deareis.tooltips as tooltips
import dearpygui.dearpygui as dpg
from deareis.gui.plots import (
    Bode,
    MuXps,
    Nyquist,
    Residuals,
)
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class ExploratoryResults:
    def __init__(
        self,
        data: DataSet,
        results: List[TestResult],
        settings: TestSettings,
        num_RCs: ndarray,
        callback: Callable,
        state,
    ):
        assert type(data) is DataSet
        assert type(results) is list and all(
            map(lambda _: type(_) is TestResult, results)
        )
        assert type(settings) is TestSettings
        assert type(num_RCs) is ndarray
        self.state = state
        dpg.split_frame(delay=33)
        self.callback: Callable = callback
        self.window: int = dpg.generate_uuid()
        self.result_combo: int = dpg.generate_uuid()
        self.accept_button: int = dpg.generate_uuid()
        self.mu_xps_plot: Optional[MuXps] = None
        self.residuals_plot: Optional[Residuals] = None
        self.nyquist_plot: Optional[Nyquist] = None
        self.bode_plot: Optional[Bode] = None
        self.key_handler: int = dpg.generate_uuid()
        self._assemble()
        self._setup_keybindings()
        self.data: DataSet = data
        self.results: List[TestResult] = results
        self.settings: TestSettings = settings
        self.num_RCs: ndarray = num_RCs
        self.mu_crit: float = settings.mu_criterion
        default_result: TestResult = results[0]
        results.sort(key=lambda _: _.num_RC)
        default_label: str = ""
        self.label_to_result: Dict[str, TestResult] = {}
        result: TestResult
        for result in results:
            label: str = (
                f"{result.num_RC}: µ = {result.mu:.3f}, "
                + f"log X² (ps.) = {log(result.pseudo_chisqr):.3f}"
            )
            if result == default_result:
                label += " *"
                default_label = label
            self.label_to_result[label] = result
        assert default_label != ""
        self.labels: List[str] = list(self.label_to_result.keys())
        self.result_index: int = self.labels.index(default_label)
        dpg.configure_item(
            self.result_combo,
            items=self.labels,
            default_value=default_label,
        )
        self.mu: ndarray = array(list(map(lambda _: _.mu, results)))
        self.Xps: ndarray = log(array(list(map(lambda _: _.pseudo_chisqr, results))))
        assert self.mu_xps_plot is not None
        self.mu_xps_plot.plot(
            num_RCs=self.num_RCs,
            mu=self.mu,
            Xps=self.Xps,
            mu_criterion=self.mu_crit,
            num_RC=min(self.num_RCs),
        )
        self.plot(default_label)
        signals.register(Signal.VIEWPORT_RESIZED, self.resize)

    def _setup_keybindings(self):
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(
                    dpg.get_item_user_data(self.accept_button),
                    keybinding=True,
                ),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.plot(
                    self.labels[(self.result_index - 1) % len(self.labels)]
                ),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.plot(
                    self.labels[(self.result_index + 1) % len(self.labels)]
                ),
            )

    def _assemble(self):
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        with dpg.window(
            label="Exploratory test results",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            no_resize=True,
            tag=self.window,
            on_close=self.close,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Num. RC elements")
                attach_tooltip(tooltips.kramers_kronig.exploratory_result)
                dpg.add_combo(
                    width=-100,
                    tag=self.result_combo,
                )
                dpg.add_button(
                    label="Accept",
                    width=-1,
                    tag=self.accept_button,
                )
            with dpg.child_window(
                border=True,
                width=-1,
                height=-1,
            ):
                with dpg.group(horizontal=True):
                    self.mu_xps_plot = MuXps()
                    self.residuals_plot = Residuals()
                    self.residuals_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                    )
                with dpg.group(horizontal=True):
                    self.nyquist_plot = Nyquist()
                    self.nyquist_plot.plot(
                        real=array([]),
                        imaginary=array([]),
                        label="Data",
                        theme=themes.nyquist.data,
                    )
                    self.nyquist_plot.plot(
                        real=array([]),
                        imaginary=array([]),
                        label="Fit",
                        show_label=False,
                        line=True,
                        theme=themes.nyquist.simulation,
                    )
                    self.nyquist_plot.plot(
                        real=array([]),
                        imaginary=array([]),
                        label="Fit",
                        theme=themes.nyquist.simulation,
                    )
                    self.bode_plot = Bode()
                    self.bode_plot.plot(
                        frequency=array([]),
                        magnitude=array([]),
                        phase=array([]),
                        labels=(
                            "|Z| (d)",
                            "phi (d)",
                        ),
                        themes=(
                            themes.bode.magnitude_data,
                            themes.bode.phase_data,
                        ),
                    )
                    self.bode_plot.plot(
                        frequency=array([]),
                        magnitude=array([]),
                        phase=array([]),
                        labels=("|Z| (f)", "phi (f)"),
                        show_labels=False,
                        line=True,
                        themes=(
                            themes.bode.magnitude_simulation,
                            themes.bode.phase_simulation,
                        ),
                    )
                    self.bode_plot.plot(
                        frequency=array([]),
                        magnitude=array([]),
                        phase=array([]),
                        labels=("|Z| (f)", "phi (f)"),
                        themes=(
                            themes.bode.magnitude_simulation,
                            themes.bode.phase_simulation,
                        ),
                    )
        dpg.set_item_callback(self.result_combo, lambda s, a: self.plot(a))
        dpg.set_item_callback(
            self.accept_button,
            lambda s, a, u: self.accept(u),
        )
        self.resize(1, 1)

    def plot(self, label: str):
        assert self.mu_xps_plot is not None
        assert self.residuals_plot is not None
        assert self.nyquist_plot is not None
        assert self.bode_plot is not None
        self.result_index = self.labels.index(label)
        dpg.set_value(self.result_combo, label)
        # Clear plots
        self.residuals_plot.clear(delete=False)
        self.nyquist_plot.clear(delete=False)
        self.bode_plot.clear(delete=False)
        # Retrieve the chosen result
        result: TestResult = self.label_to_result[label]
        dpg.set_item_user_data(self.accept_button, result)
        # Mu-Xps vs num RC
        self.mu_xps_plot.update(
            num_RC=result.num_RC,
        )
        # Residuals
        freq: ndarray
        real: ndarray
        imag: ndarray
        freq, real, imag = result.get_residual_data()
        self.residuals_plot.update(
            index=0,
            frequency=freq,
            real=real,
            imaginary=imag,
        )
        # Data and fit
        # - Nyquist
        real, imag = self.data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        real, imag = result.get_nyquist_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        real, imag = result.get_nyquist_data()
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        # Bode
        mag: ndarray
        phase: ndarray
        freq, mag, phase = self.data.get_bode_data()
        self.bode_plot.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = result.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = result.get_bode_data()
        self.bode_plot.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        dpg.split_frame()
        self.mu_xps_plot.queue_limits_adjustment()
        self.residuals_plot.queue_limits_adjustment()
        self.nyquist_plot.queue_limits_adjustment()
        self.bode_plot.queue_limits_adjustment()

    def resize(self, width: int, height: int):
        assert self.mu_xps_plot is not None
        assert self.residuals_plot is not None
        assert self.nyquist_plot is not None
        assert self.bode_plot is not None
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        dpg.split_frame()
        width, height = dpg.get_item_rect_size(self.window)
        width -= 56
        height -= 80
        width = int(floor(width / 2))
        height = int(floor(height / 2))
        self.mu_xps_plot.resize(width, height)
        self.residuals_plot.resize(width, height)
        self.nyquist_plot.resize(width, height)
        self.bode_plot.resize(width, height)

    def close(self):
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        signals.unregister(Signal.VIEWPORT_RESIZED, self.resize)

    def accept(self, result: TestResult, keybinding: bool = False):
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        self.callback(self.data, result, self.settings)
        self.close()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    List,
    Optional,
)
from numpy import (
    array,
    ceil,
    floor,
    log10 as log,
    ndarray,
)
import dearpygui.dearpygui as dpg
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class DRT(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=kwargs.get("legend_horizontal", True),
                location=kwargs.get("legend_location", dpg.mvPlot_Location_North),
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="tau (s)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="gamma (ohm)",
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return len(dpg.get_item_children(self._y_axis, slot=1)) == 0

    def clear(self, *args, **kwargs):
        delete: bool = kwargs.get("delete", True)
        if delete:
            dpg.delete_item(self._y_axis, children_only=True)
            self._series.clear()
        else:
            i: int
            series: int
            for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
                self._series[i]["tau"] = array([])
                self._series[i]["gamma"] = array([])
                dpg.set_value(series, [[], []])

    def delete_series(self, *args, **kwargs):
        index: int = kwargs.get("index", -1)
        from_index: int = kwargs.get("from_index", -1)
        i: int
        series: int
        if index >= 0:
            for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
                if i != index:
                    continue
                dpg.delete_item(series)
                self._series.pop(i)
        elif from_index >= 0:
            for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
                if i < from_index:
                    continue
                dpg.delete_item(series)
                self._series[i] = None
            while None in self._series:
                self._series.remove(None)
        else:
            raise Exception("Unsupported case!")

    def update(self, index: int, *args, **kwargs):
        assert type(index) is int and index >= 0, index
        assert len(self._series) > index, (
            index,
            len(self._series),
        )
        assert len(args) == 0, args
        tau: ndarray = kwargs["tau"]
        gamma: ndarray = kwargs["gamma"]
        assert type(tau) is ndarray, tau
        assert type(gamma) is ndarray, gamma
        assert tau.shape == gamma.shape, (
            tau.shape,
            gamma.shape,
        )
        i: int
        series: int
        for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
            if i != index:
                continue
            dpg.set_value(series, [list(tau), list(gamma)])
            if "label" in kwargs:
                label: Optional[str] = kwargs.get("label")
                assert type(label) is str or label is None, label
                dpg.set_item_label(series, label)
            self._series[index].update(kwargs)
            break

    def plot(self, *args, **kwargs) -> int:
        assert len(args) == 0, args
        tau: ndarray = kwargs["tau"]
        gamma: Optional[ndarray] = kwargs.get(
            "gamma",
            kwargs.get(
                "imaginary",
                kwargs.get("mean"),
            ),
        )
        lower: Optional[ndarray] = kwargs.get("lower")
        upper: Optional[ndarray] = kwargs.get("upper")
        label: str = kwargs.get("label", "")
        line: bool = kwargs.get("line", False)
        theme: Optional[int] = kwargs.get("theme")
        show_label: bool = kwargs.get("show_label", True)
        assert type(tau) is ndarray, tau
        assert type(gamma) is ndarray or gamma is None, gamma
        assert type(lower) is ndarray or lower is None, lower
        assert type(upper) is ndarray or upper is None, upper
        assert type(label) is str, label
        assert type(line) is bool, line
        assert type(theme) is int or theme is None, theme
        assert type(show_label) is bool, show_label
        tag: int
        if gamma is not None:
            tag = dpg.add_line_series(
                x=list(tau),
                y=list(gamma),
                label=label if show_label else None,
                parent=self._y_axis,
            )
        elif lower is not None and upper is not None:
            tag = dpg.add_shade_series(
                x=list(tau),
                y1=list(lower),  # Lower bounds
                y2=list(upper),  # Upper bounds
                label=label if show_label else None,
                parent=self._y_axis,
            )
        else:
            raise Exception("No data to plot!")
        self._series.append(kwargs)
        if theme is not None:
            dpg.bind_item_theme(tag, theme)
        return tag

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        x_min: Optional[float] = None
        x_max: Optional[float] = None
        y_min: Optional[float] = None
        y_max: Optional[float] = None
        for kwargs in self._series:
            tau: ndarray = kwargs["tau"]
            if tau.size > 0:
                if x_min is None or min(tau) < x_min:
                    x_min = min(tau)
                if x_max is None or max(tau) > x_max:
                    x_max = max(tau)
            gamma: Optional[ndarray] = kwargs.get(
                "gamma",
                kwargs.get(
                    "imaginary",
                    kwargs.get("mean"),
                ),
            )
            if gamma is not None:
                if gamma.size > 0:
                    if y_min is None or min(gamma) < y_min:
                        y_min = min(gamma)
                    if y_max is None or max(gamma) > y_max:
                        y_max = max(gamma)
            else:
                lower: ndarray = kwargs["lower"]
                upper: ndarray = kwargs["upper"]
                if lower.size > 0 and upper.size > 0:
                    if y_min is None or min(lower) < y_min:
                        y_min = min(lower)
                    if y_max is None or max(upper) > y_max:
                        y_max = max(upper)
        if x_min is None:
            x_min = 0.5
            x_max = 1.0
            y_min = 0.0
            y_max = 1.0
        else:
            dx: float = 0.1
            x_min = 10 ** (floor(log(x_min) / dx) * dx - dx)
            x_max = 10 ** (ceil(log(x_max) / dx) * dx + dx)
            dy: float = abs(y_max - y_min) * 0.05
            if dy < 1e-7:
                y_min = -0.5
                y_max = 0.5
            else:
                y_min = y_min - dy
                y_max = y_max + dy
        dpg.split_frame()
        dpg.set_axis_limits(self._x_axis, ymin=x_min, ymax=x_max)
        dpg.set_axis_limits(self._y_axis, ymin=y_min, ymax=y_max)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis,
            ],
            [
                self._x_axis,
                self._y_axis,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from multiprocessing import cpu_count
from traceback import format_exc
from typing import (
    List,
    Optional,
)
from numpy import (
    array,
    ndarray,
)
from pyimpspec import FittingError
import deareis.api.kramers_kronig as api
from deareis.data import (
    DataSet,
    PlotSettings,
    Project,
    TestResult,
    TestSettings,
)
from deareis.enums import (
    TestMode,
)
from deareis.gui import ProjectTab
from deareis.gui.kramers_kronig.exploratory_results import ExploratoryResults
from deareis.signals import Signal
import deareis.signals as signals
from deareis.state import STATE


def select_test_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    test: Optional[TestResult] = kwargs.get("test")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or test is None:
        return
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading test result")
    project_tab.select_test_result(test, data)
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def delete_test_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    test: Optional[TestResult] = kwargs.get("test")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or test is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting test result")
    settings: Optional[PlotSettings] = project_tab.get_active_plot()
    update_plot: bool = (
        test.uuid in settings.series_order if settings is not None else False
    )
    project.delete_test(
        data=data,
        test=test,
    )
    project_tab.populate_tests(project, data)
    if settings is not None:
        project_tab.plotting_tab.populate_tests(
            project.get_all_tests(),
            project.get_data_sets(),
            settings,
        )
    if update_plot:
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def apply_test_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[TestSettings] = kwargs.get("settings")
    if settings is None:
        return
    project_tab.set_test_settings(settings)


def accept_exploratory_result(data: DataSet, test: TestResult, settings: TestSettings):
    assert type(data) is DataSet
    assert type(test) is TestResult
    assert type(settings) is TestSettings
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    project.add_test(
        data=data,
        test=test,
    )
    project_tab.populate_tests(project, data)
    project_tab.plotting_tab.populate_tests(
        project.get_all_tests(),
        project.get_data_sets(),
        project_tab.get_active_plot(),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def show_exploratory_results(
    data: DataSet,
    results: List[TestResult],
    settings: TestSettings,
    num_RCs: ndarray,
):
    assert type(data) is DataSet
    assert type(results) is list and all(map(lambda _: type(_) is TestResult, results))
    assert type(settings) is TestSettings
    assert type(num_RCs) is ndarray
    exploratory_results: ExploratoryResults = ExploratoryResults(
        data=data,
        results=results,
        settings=settings,
        num_RCs=num_RCs,
        callback=accept_exploratory_result,
        state=STATE,
    )
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=exploratory_results.window,
        window_object=exploratory_results,
    )


def perform_test(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    data: Optional[DataSet] = kwargs.get("data")
    settings: Optional[TestSettings] = kwargs.get("settings")
    if data is None or settings is None:
        return
    assert data.get_num_points() > 0, "There are no data points to test!"
    # Prevent the GUI from becoming unresponsive or sluggish
    num_procs: int = max(2, cpu_count() - 1)
    if settings.mode == TestMode.AUTO or settings.mode == TestMode.MANUAL:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Performing test(s)")
        try:
            test: TestResult = api.perform_test(
                data=data,
                settings=settings,
                num_procs=num_procs,
            )
        except FittingError:
            signals.emit(Signal.SHOW_ERROR_MESSAGE, traceback=format_exc())
            return
        signals.emit(Signal.HIDE_BUSY_MESSAGE)
        project.add_test(
            data=data,
            test=test,
        )
        project_tab.populate_tests(project, data)
        project_tab.plotting_tab.populate_tests(
            project.get_all_tests(),
            project.get_data_sets(),
            project_tab.get_active_plot(),
        )
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    elif settings.mode == TestMode.EXPLORATORY:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Performing test")
        try:
            results: List[TestResult] = api.perform_exploratory_tests(
                data=data,
                settings=settings,
                num_procs=num_procs,
            )
        except FittingError:
            signals.emit(Signal.SHOW_ERROR_MESSAGE, traceback=format_exc())
            return
        signals.emit(Signal.HIDE_BUSY_MESSAGE)
        num_RCs: ndarray = array(list(range(1, settings.num_RC + 1)))
        show_exploratory_results(
            data,
            results,
            settings,
            num_RCs,
        )
    else:
        raise Exception("Unsupported mode!")



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
import dearpygui.dearpygui as dpg
from numpy import (
    array,
    integer,
    issubdtype,
    ndarray,
)
from deareis.data import DataSet
from deareis.data.kramers_kronig import TestResult
from deareis.data.drt import DRTResult
from deareis.data.fitting import FitResult
from deareis.data.simulation import SimulationResult
from deareis.enums import (
    PLOT_EXTENSIONS,
    PlotLegendLocation,
    PlotPreviewLimit,
    PlotType,
    PlotUnits,
)
from deareis.themes import (
    create_plot_series_theme,
    get_random_color_marker,
    update_plot_series_theme_color,
    update_plot_series_theme_marker,
)


@dataclass(frozen=True)
class PlotSeries:
    """
    A class that represents the data used to plot an item/series.
    """

    data: Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
    label: str
    color: Tuple[float, float, float, float]
    marker: int
    line: bool
    legend: bool

    def __repr__(self) -> str:
        return f"PlotSeries ({self.get_label()}, {hex(id(self))})"

    def get_label(self) -> str:
        return self.label

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        if type(self.data) is DataSet or type(self.data) is DRTResult:
            return self.data.get_frequency()
        return self.data.get_frequency(num_per_decade=num_per_decade)

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        if type(self.data) is DataSet or type(self.data) is DRTResult:
            return self.data.get_impedance()
        return self.data.get_impedance(num_per_decade=num_per_decade)

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        if type(self.data) is DataSet or type(self.data) is DRTResult:
            return self.data.get_nyquist_data()
        return self.data.get_nyquist_data(num_per_decade=num_per_decade)

    def get_bode_data(
        self,
        num_per_decade: int = -1,
    ) -> Tuple[ndarray, ndarray, ndarray]:
        if type(self.data) is DataSet or type(self.data) is DRTResult:
            return self.data.get_bode_data()
        return self.data.get_bode_data(num_per_decade=num_per_decade)

    def get_tau(self) -> ndarray:
        if type(self.data) is not DRTResult:
            return array([])
        return self.data.get_tau()

    def get_gamma(self, imaginary: bool = False) -> ndarray:
        if type(self.data) is not DRTResult:
            return array([])
        return self.data.get_gamma(imaginary=imaginary)

    def get_drt_data(self, imaginary: bool = False) -> Tuple[ndarray, ndarray]:
        if type(self.data) is not DRTResult:
            return (
                array([]),
                array([]),
            )
        return self.data.get_drt_data(imaginary=imaginary)

    def get_drt_credible_intervals(self) -> Tuple[ndarray, ndarray, ndarray, ndarray]:
        if type(self.data) is not DRTResult:
            return (
                array([]),
                array([]),
                array([]),
                array([]),
            )
        return self.data.get_drt_credible_intervals()

    def get_color(self) -> Tuple[float, float, float, float]:
        return self.color

    def get_marker(self) -> int:
        return self.marker

    def has_markers(self) -> bool:
        return self.marker >= 0

    def has_line(self) -> bool:
        return self.line

    def has_legend(self) -> bool:
        return self.legend


VERSION: int = 1


def _parse_export_settings_v1(dictionary: dict) -> dict:
    # TODO: Updated implementation when VERSION is incremented
    return dictionary


@dataclass(frozen=True)
class PlotExportSettings:
    units: PlotUnits
    width: float
    height: float
    dpi: int
    preview_limit: PlotPreviewLimit
    show_title: bool
    show_legend: bool
    legend_location: PlotLegendLocation
    show_grid: bool
    has_tight_layout: bool
    num_per_decade: int
    extension: str
    clear_registry: bool
    disable_preview: bool

    @classmethod
    def from_dict(Class, dictionary: dict) -> "PlotExportSettings":
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_export_settings_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        del dictionary["version"]
        dictionary = parsers[version](dictionary)
        dictionary["units"] = PlotUnits(dictionary["units"])
        dictionary["preview_limit"] = PlotPreviewLimit(dictionary["preview_limit"])
        dictionary["legend_location"] = PlotLegendLocation(
            dictionary["legend_location"]
        )
        if dictionary["extension"] not in PLOT_EXTENSIONS:
            dictionary["extension"] = ".png"
        return Class(**dictionary)

    def to_dict(self) -> dict:
        return {
            "version": VERSION,
            "units": int(self.units),
            "width": self.width,
            "height": self.height,
            "dpi": self.dpi,
            "preview_limit": int(self.preview_limit),
            "show_title": self.show_title,
            "show_legend": self.show_legend,
            "legend_location": int(self.legend_location),
            "show_grid": self.show_grid,
            "has_tight_layout": self.has_tight_layout,
            "num_per_decade": self.num_per_decade,
            "extension": self.extension,
            "clear_registry": self.clear_registry,
            "disable_preview": self.disable_preview,
        }


def _parse_settings_v1(dictionary: dict) -> dict:
    # TODO: Implement when VERSION has been incremented
    return dictionary


# TODO: Make it possible to (re)store the limits of plots (e.g. from session to session)?
@dataclass
class PlotSettings:
    """
    A class representing a complex plot that can contain one or more data sets, Kramers-Kronig test results, DRT analysis results, equivalent circuit fitting results, and simulation results.
    """

    plot_label: str
    plot_type: PlotType
    series_order: List[str]  # UUID
    labels: Dict[str, str]  # UUID: label
    colors: Dict[str, List[float]]  # UUID: RGBA
    markers: Dict[str, int]  # UUID: enum
    # TODO: Add a toggle for filled markers
    # filled_markers: Dict[str, bool]
    # TODO: Replace show_lines with lines (UUID: enum)?
    # lines: Dict[str, int]
    show_lines: Dict[str, bool]  # UUID: flag
    themes: Dict[str, int]  # UUID: DPG UUID
    uuid: str

    def __eq__(self, other) -> bool:
        try:
            assert isinstance(other, type(self)), other
            assert self.uuid == other.uuid, (
                self.uuid,
                other.uuid,
            )
            assert self.get_label() == other.get_label(), (
                self.get_label(),
                other.get_label(),
            )
            assert self.get_type() == other.get_type(), (
                self.get_type(),
                other.get_type(),
            )
            assert ",".join(self.series_order) == ",".join(other.series_order), (
                self.series_order,
                other.series_order,
            )
            key: str
            for key in self.labels:
                assert key in other.labels
                assert self.labels[key] == other.labels[key]
            for key in self.colors:
                assert key in other.colors
                assert self.colors[key] == other.colors[key]
            for key in self.markers:
                assert key in other.markers
                assert self.markers[key] == other.markers[key]
            for key in self.show_lines:
                assert key in other.show_lines
                assert self.show_lines[key] == other.show_lines[key]
        except AssertionError:
            return False
        return True

    def __repr__(self) -> str:
        return f"PlotSettings ({self.get_label()}, {hex(id(self))})"

    def recreate_themes(self):
        uuid: str
        for uuid in list(self.themes.keys()):
            item: int = self.themes[uuid]
            if item < 0:
                self.themes[uuid] = dpg.generate_uuid()
                create_plot_series_theme(
                    self.colors[uuid],
                    self.markers[uuid] if self.markers[uuid] >= 0 else 0,
                    self.themes[uuid],
                )
            else:
                if not dpg.does_item_exist(self.themes[uuid]):
                    del self.labels[uuid]
                    del self.colors[uuid]
                    del self.markers[uuid]
                    del self.show_lines[uuid]
                    del self.themes[uuid]
                else:
                    update_plot_series_theme_color(self.themes[uuid], self.colors[uuid])
                    update_plot_series_theme_marker(
                        self.themes[uuid], self.markers[uuid]
                    )

    @classmethod
    def from_dict(Class, dictionary: dict) -> "PlotSettings":
        assert type(dictionary) is dict, dictionary
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_settings_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        del dictionary["version"]
        dictionary = parsers[version](dictionary)
        dictionary["plot_type"] = PlotType(dictionary["plot_type"])
        settings = Class(**dictionary)
        settings.recreate_themes()
        return settings

    def to_dict(self, session: bool) -> dict:
        return {
            "uuid": self.uuid,
            "version": VERSION,
            "plot_label": self.plot_label,
            "plot_type": self.plot_type,
            "series_order": self.series_order.copy(),
            "labels": self.labels.copy(),
            "colors": {k: v[:] for k, v in self.colors.items()},
            "markers": self.markers.copy(),
            "show_lines": self.show_lines.copy(),
            "themes": self.themes.copy() if session else {k: -1 for k in self.themes},
        }

    def get_label(self) -> str:
        return self.plot_label

    def set_label(self, label: str):
        if label.strip() == "":
            return
        self.plot_label = label

    def get_type(self) -> PlotType:
        return self.plot_type

    def set_type(self, plot_type: PlotType):
        self.plot_type = plot_type

    def get_series_label(self, uuid: str) -> str:
        assert type(uuid) is str, uuid
        return self.labels.get(uuid, "")

    def set_series_label(self, uuid: str, label: str):
        assert type(uuid) is str, uuid
        assert type(label) is str, label
        self.labels[uuid] = label

    def get_series_theme(self, uuid: str) -> int:
        assert type(uuid) is str, uuid
        return self.themes.get(uuid, -1)

    def get_series_color(self, uuid: str) -> List[float]:
        assert type(uuid) is str, uuid
        color: Optional[List[float]] = self.colors.get(uuid)
        if color is None:
            return [255.0, 255.0, 255.0, 255.0]
        return color

    def set_series_color(self, uuid: str, color: List[float]):
        assert type(uuid) is str, uuid
        if type(color) is tuple:
            color = list(color)
        assert type(color) is list and len(color) == 4, color
        theme: int = self.themes[uuid]
        update_plot_series_theme_color(theme, color)
        self.colors[uuid] = color

    def get_series_marker(self, uuid: str) -> int:
        assert type(uuid) is str, uuid
        return self.markers.get(uuid, -1)

    def set_series_marker(self, uuid: str, marker: int):
        assert type(uuid) is str, uuid
        assert issubdtype(type(marker), integer), marker
        theme: int = self.themes[uuid]
        if marker >= 0:
            update_plot_series_theme_marker(theme, marker)
        self.markers[uuid] = marker

    def get_series_line(self, uuid: str) -> bool:
        assert type(uuid) is str, uuid
        return self.show_lines.get(uuid, False)

    def set_series_line(self, uuid: str, state: bool):
        assert type(uuid) is str, uuid
        assert type(state) is bool, state
        self.show_lines[uuid] = state

    def add_series(
        self,
        series: Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult],
    ):
        # TODO: Refactor so that series is replaced by uuid?
        # Include the type as another argument to determine whether or not a line should be drawn?
        assert (
            type(series) is DataSet
            or type(series) is TestResult
            or type(series) is DRTResult
            or type(series) is FitResult
            or type(series) is SimulationResult
        ), series
        uuid: str = series.uuid
        if uuid in self.series_order:
            return
        if uuid not in self.themes:
            self.labels[uuid] = ""
            color: List[float]
            marker: int
            color, marker = get_random_color_marker(self.themes)
            self.colors[uuid] = color
            self.markers[uuid] = marker
            self.themes[uuid] = dpg.generate_uuid()
            create_plot_series_theme(
                self.colors[uuid],
                self.markers[uuid] if self.markers[uuid] >= 0 else 0,
                self.themes[uuid],
            )
            self.show_lines[uuid] = type(series) is not DataSet
        self.series_order.append(uuid)

    def remove_series(self, uuid: str):
        assert type(uuid) is str, uuid
        if uuid not in self.series_order:
            return
        self.series_order.remove(uuid)

    def find_series(
        self,
        uuid: str,
        datasets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
    ) -> Optional[Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]]:
        def find_dataset() -> Optional[DataSet]:
            data: DataSet
            for data in datasets:
                if data.uuid == uuid:
                    return data
            return None

        def find_test() -> Optional[TestResult]:
            test: TestResult
            for test in [test for _ in tests.values() for test in _]:
                if test.uuid == uuid:
                    return test
            return None

        def find_drt() -> Optional[DRTResult]:
            drt: DRTResult
            for drt in [drt for _ in drts.values() for drt in _]:
                if drt.uuid == uuid:
                    return drt
            return None

        def find_fit() -> Optional[FitResult]:
            fit: FitResult
            for fit in [fit for _ in fits.values() for fit in _]:
                if fit.uuid == uuid:
                    return fit
            return None

        def find_simulation() -> Optional[SimulationResult]:
            sim: SimulationResult
            for sim in simulations:
                if sim.uuid == uuid:
                    return sim
            return None

        data: Optional[DataSet] = find_dataset()
        if data is not None:
            return data
        test: Optional[TestResult] = find_test()
        if test is not None:
            return test
        drt: Optional[DRTResult] = find_drt()
        if drt is not None:
            return drt
        fit: Optional[FitResult] = find_fit()
        if fit is not None:
            return fit
        return find_simulation()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
import dearpygui.dearpygui as dpg
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from numpy import (
    floating,
    integer,
    issubdtype,
    ndarray,
)
from pyimpspec.plot import mpl
from deareis.data import (
    DRTResult,
    DataSet,
    FitResult,
    PlotSettings,
    Project,
    SimulationResult,
    TestResult,
)
from deareis.enums import PlotType


MPL_MARKERS: Dict[int, str] = {
    dpg.mvPlotMarker_Down: "v",
    dpg.mvPlotMarker_Left: "<",
    dpg.mvPlotMarker_Right: ">",
    dpg.mvPlotMarker_Up: "^",
    dpg.mvPlotMarker_Asterisk: "*",
    dpg.mvPlotMarker_Circle: "o",
    dpg.mvPlotMarker_Cross: "x",
    dpg.mvPlotMarker_Diamond: "D",
    dpg.mvPlotMarker_Plus: "+",
    dpg.mvPlotMarker_Square: "s",
}

_UNFILLED_MARKERS: List[int] = [
    dpg.mvPlotMarker_Cross,
    dpg.mvPlotMarker_Plus,
]


def plot(
    settings: PlotSettings,
    project: Project,
    x_limits: Optional[Tuple[Optional[float], Optional[float]]] = None,
    y_limits: Optional[Tuple[Optional[float], Optional[float]]] = None,
    show_title: bool = True,
    show_legend: Optional[bool] = None,
    legend_loc: Union[int, str] = 0,
    show_grid: bool = False,
    tight_layout: bool = False,
    fig: Optional[Figure] = None,
    axis: Optional[Axes] = None,
    num_per_decade: int = 100,
) -> Tuple[Figure, Axes]:
    """
    Plot a complex plot containing one or more items from a project based on the provided settings.

    Parameters
    ----------
    settings: PlotSettings
        The settings for the plot.

    project: Project
        The project that the plot is a part of.

    x_limits: Optional[Tuple[Optional[float], Optional[float]]] = None
        The lower and upper limits of the x-axis.

    y_limits: Optional[Tuple[Optional[float], Optional[float]]] = None
        The lower and upper limits of the y-axis.

    show_title: bool = True
        Whether or not to include the title in the figure.

    show_legend: Optional[bool] = None
        Whether or not to include a legend in the figure.

    legend_loc: Union[int, str] = 0
        The position of the legend in the figure. See matplotlib's documentation for valid values.

    show_grid: bool = False
        Whether or not to include a grid in the figure.

    tight_layout: bool = False
        Whether or not to apply a tight layout that the sizes of the reduces margins.

    fig: Optional[Figure] = None
        The matplotlib.figure.Figure instance to use when plotting the data.

    axis: Optional[Axes] = None
        The matplotlib.axes.Axes instance to use when plotting the data.

    num_per_decade: int = 100
        If any circuit fits, circuit simulations, or Kramers-Kronig test results are included in the plot, then this parameter can be used to change how many points are used to draw the line (i.e. how smooth or angular the line looks).
    """
    assert type(settings) is PlotSettings, settings
    assert type(project) is Project, project
    assert x_limits is None or (
        type(x_limits) is tuple
        and len(x_limits) == 2
        and all(
            map(
                lambda _: _ is None
                or issubdtype(type(_), integer)
                or issubdtype(type(_), floating),
                x_limits,
            )
        )
    ), x_limits
    assert y_limits is None or (
        type(y_limits) is tuple
        and len(y_limits) == 2
        and all(
            map(
                lambda _: _ is None
                or issubdtype(type(_), integer)
                or issubdtype(type(_), floating),
                y_limits,
            )
        )
    ), y_limits
    assert type(show_title) is bool, show_title
    assert type(show_legend) is bool or show_legend is None, show_legend
    assert issubdtype(type(legend_loc), integer) or type(legend_loc) is str, legend_loc
    assert type(show_grid) is bool, show_grid
    assert type(tight_layout) is bool, tight_layout
    assert type(fig) is Figure or fig is None
    if fig is None:
        fig, axis = plt.subplots()
    assert axis is not None
    assert (
        issubdtype(type(num_per_decade), integer) and num_per_decade >= 1
    ), num_per_decade
    plot_type: PlotType = settings.get_type()
    selected_series: List[
        Tuple[str, Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]]
    ] = []
    uuid: str
    for uuid in settings.series_order:
        series: Optional[
            Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
        ]
        series = settings.find_series(
            uuid,
            project.get_data_sets(),
            project.get_all_tests(),
            project.get_all_drts(),
            project.get_all_fits(),
            project.get_simulations(),
        )
        if series is None:
            continue
        # Filter out stuff that cannot be plotted
        if plot_type == PlotType.DRT and type(series) is not DRTResult:
            continue
        selected_series.append(
            (
                uuid,
                series,
            )
        )
    num_series: int = len(selected_series)
    for i, (uuid, series) in enumerate(selected_series):
        label: Optional[str] = settings.get_series_label(uuid) or series.get_label()
        if label.strip() == "" and label != "":  # type: ignore
            label = ""
        if label is not None and show_legend is None:
            show_legend = True
        color: List[float] = list(
            map(lambda _: _ / 255.0, settings.get_series_color(uuid))
        )
        has_line: bool = settings.get_series_line(uuid)
        marker: Optional[str] = MPL_MARKERS.get(settings.get_series_marker(uuid))
        scatter_data: List[ndarray]
        line_data: List[ndarray]
        if plot_type == PlotType.NYQUIST:
            if has_line:
                mpl.plot_nyquist(
                    series,
                    color=color,
                    marker=marker,
                    line=True,
                    label=label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=num_per_decade,
                    adjust_axes=i == num_series - 1,
                )
            if marker is not None:
                mpl.plot_nyquist(
                    series,
                    color=color,
                    marker=marker,
                    line=False,
                    label="" if has_line else label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=-1,
                    adjust_axes=i == num_series - 1,
                )
        elif plot_type == PlotType.BODE_MAGNITUDE:
            if has_line:
                mpl.plot_impedance_magnitude(
                    series,
                    color=color,
                    marker=marker,
                    line=True,
                    label=label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=num_per_decade,
                    adjust_axes=i == num_series - 1,
                )
            if marker is not None:
                mpl.plot_impedance_magnitude(
                    series,
                    color=color,
                    marker=marker,
                    line=False,
                    label="" if has_line else label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=-1,
                    adjust_axes=i == num_series - 1,
                )
        elif plot_type == PlotType.BODE_PHASE:
            if has_line:
                mpl.plot_impedance_phase(
                    series,
                    color=color,
                    marker=marker,
                    line=True,
                    label=label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=num_per_decade,
                    adjust_axes=i == num_series - 1,
                )
            if marker is not None:
                mpl.plot_impedance_phase(
                    series,
                    color=color,
                    marker=marker,
                    line=False,
                    label="" if has_line else label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=-1,
                    adjust_axes=i == num_series - 1,
                )
        elif plot_type == PlotType.DRT:
            mpl.plot_gamma(
                series,
                color=color,
                label=label,
                legend=False,
                fig=fig,
                axis=axis,
                adjust_axes=i == num_series - 1,
            )
        elif plot_type == PlotType.IMPEDANCE_REAL:
            if has_line:
                mpl.plot_real_impedance(
                    series,
                    color=color,
                    marker=marker,
                    line=True,
                    label=label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=num_per_decade,
                    adjust_axes=i == num_series - 1,
                )
            if marker is not None:
                mpl.plot_real_impedance(
                    series,
                    color=color,
                    marker=marker,
                    line=False,
                    label="" if has_line else label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=-1,
                    adjust_axes=i == num_series - 1,
                )
        elif plot_type == PlotType.IMPEDANCE_IMAGINARY:
            if has_line:
                mpl.plot_imaginary_impedance(
                    series,
                    color=color,
                    marker=marker,
                    line=True,
                    label=label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=num_per_decade,
                    adjust_axes=i == num_series - 1,
                )
            if marker is not None:
                mpl.plot_imaginary_impedance(
                    series,
                    color=color,
                    marker=marker,
                    line=False,
                    label="" if has_line else label,
                    legend=False,
                    fig=fig,
                    axis=axis,
                    num_per_decade=-1,
                    adjust_axes=i == num_series - 1,
                )
        else:
            raise Exception(f"Unsupported plot type: {plot_type=}")
    if x_limits is not None:
        axis.set_xlim(x_limits)
    if y_limits is not None:
        axis.set_ylim(y_limits)
    if show_title and settings.get_label() != "":
        fig.suptitle(settings.get_label())
    if show_legend:
        axis.legend(loc=legend_loc)
    axis.grid(visible=show_grid)
    if tight_layout:
        fig.tight_layout()
    return (
        fig,
        axis,
    )
